/* -*- Mode:C++; c-file-style:"gnu"; indent-tabs-mode:nil; -*- */
/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "ns3/core-module.h"
#include "ns3/simulator-module.h"
#include "ns3/node-module.h"
#include "ns3/helper-module.h"
#include "ns3/wifi-module.h"
#include "ns3/mobility-module.h"
#include <fstream>

// Default Network Topology
//
//   Wifi 10.1.3.0
//                 AP   
//  *    *    *    *
//  |    |    |    |    10.1.1.0
// n5   n6   n7   n0 -------------- n1   n2   n3   n4
//                   point-to-point  |    |    |    |
//                                   ================
//                                     LAN 10.1.2.0

using namespace ns3;

NS_LOG_COMPONENT_DEFINE ("ThirdScriptExample");

int 
main (int argc, char *argv[])
{
  bool verbose = true;
//  uint32_t nCsma = 3;
  uint32_t nWifi = 2;
  uint32_t nP2p = 3;
  CommandLine cmd;
//  cmd.AddValue ("nCsma", "Number of \"extra\" CSMA nodes/devices", nCsma);
  cmd.AddValue ("nWifi", "Number of wifi STA devices", nWifi);
  cmd.AddValue ("verbose", "Tell echo applications to log if true", verbose);
  cmd.Parse (argc,argv);

  if (verbose)
    {
      LogComponentEnable("UdpEchoClientApplication", LOG_LEVEL_INFO);
      LogComponentEnable("UdpEchoServerApplication", LOG_LEVEL_INFO);
    }

  NodeContainer nC;
  nC.Create (nP2p);

  PointToPointHelper pointToPoint;
  pointToPoint.SetDeviceAttribute ("DataRate", StringValue ("5Mbps"));
  pointToPoint.SetChannelAttribute ("Delay", StringValue ("2ms"));

  NodeContainer p2p01 = NodeContainer (nC.Get(0), nC.Get (1));
  NodeContainer p2p02 = NodeContainer (nC.Get(0), nC.Get (2));
  NodeContainer p2p03 = NodeContainer (nC.Get(1), nC.Get (2));

  NetDeviceContainer p2pDevices01 = pointToPoint.Install (p2p01);
  NetDeviceContainer p2pDevices02 = pointToPoint.Install (p2p02);
  NetDeviceContainer p2pDevices03 = pointToPoint.Install (p2p03);

  NodeContainer wifiStaNodes01, wifiStaNodes02;
  wifiStaNodes01.Create (nWifi);
  wifiStaNodes02.Create (nWifi);
  NodeContainer wifiApNode = NodeContainer (nC.Get (0), nC.Get (1), nC.Get (2));

  YansWifiChannelHelper channel = YansWifiChannelHelper::Default ();
  YansWifiPhyHelper phy = YansWifiPhyHelper::Default ();
  phy.SetChannel (channel.Create ());

  WifiHelper wifi = WifiHelper::Default ();
  wifi.SetRemoteStationManager ("ns3::AarfWifiManager");

  NqosWifiMacHelper mac = NqosWifiMacHelper::Default ();

  Ssid ssid = Ssid ("ns-3-ssid");
  mac.SetType ("ns3::NqstaWifiMac", 
    "Ssid", SsidValue (ssid),
    "ActiveProbing", BooleanValue (false));

  NetDeviceContainer staDevices01, staDevices02;
  staDevices01 = wifi.Install (phy, mac, wifiStaNodes01);
  staDevices02 = wifi.Install (phy, mac, wifiStaNodes02);

  mac.SetType ("ns3::NqapWifiMac", 
    "Ssid", SsidValue (ssid),
    "BeaconGeneration", BooleanValue (true),
    "BeaconInterval", TimeValue (Seconds (2.5)));

  NetDeviceContainer apDevices;
  apDevices = wifi.Install (phy, mac, wifiApNode);

  MobilityHelper mobility;
  Ptr<ListPositionAllocator> positionAllocAp01 = CreateObject<ListPositionAllocator> ();
  Ptr<ListPositionAllocator> positionAllocAp02 = CreateObject<ListPositionAllocator> ();
  Ptr<ListPositionAllocator> positionAllocAp03 = CreateObject<ListPositionAllocator> ();
  Ptr<ListPositionAllocator> positionAllocSta01 = CreateObject<ListPositionAllocator> ();
  Ptr<ListPositionAllocator> positionAllocSta02 = CreateObject<ListPositionAllocator> ();
  Ptr<ListPositionAllocator> positionAllocSta03 = CreateObject<ListPositionAllocator> ();
  Ptr<ListPositionAllocator> positionAllocSta04 = CreateObject<ListPositionAllocator> ();

  positionAllocAp01->Add (Vector (200, 200, 0.0));
  mobility.SetPositionAllocator (positionAllocAp01);
  // setup the AP.
  mobility.SetMobilityModel ("ns3::RandomWalk2dMobilityModel",
  			 "Mode", StringValue ("Time"),
			 "Time", StringValue ("2s"),
			 "Speed", StringValue ("Constant:0.0"),
			 "Bounds", RectangleValue (Rectangle (0, 400.0, 0.0, 400.0)));
  mobility.Install (wifiApNode.Get (0));

  positionAllocAp02->Add (Vector (300, 100, 0.0));      
  mobility.SetPositionAllocator (positionAllocAp02);
  // setup the AP.
  mobility.SetMobilityModel ("ns3::RandomWalk2dMobilityModel",
  			 "Mode", StringValue ("Time"),
			 "Time", StringValue ("2s"),
			 "Speed", StringValue ("Constant:0.0"),
			 "Bounds", RectangleValue (Rectangle (0, 400.0, 0.0, 400.0)));
  mobility.Install (wifiApNode.Get (1));

  positionAllocAp03->Add (Vector (300, 300, 0.0));      
  mobility.SetPositionAllocator (positionAllocAp03);
  // setup the AP.
  mobility.SetMobilityModel ("ns3::RandomWalk2dMobilityModel",
  			 "Mode", StringValue ("Time"),
			 "Time", StringValue ("2s"),
			 "Speed", StringValue ("Constant:0.0"),
			 "Bounds", RectangleValue (Rectangle (0, 400.0, 0.0, 400.0)));
  mobility.Install (wifiApNode.Get (2));

  positionAllocSta01->Add (Vector (170, 230, 0.0));
  mobility.SetPositionAllocator (positionAllocSta01);
      
  mobility.SetMobilityModel ("ns3::RandomWalk2dMobilityModel",
    			 "Mode", StringValue ("Time"),
			 "Time", StringValue ("2s"),
			 "Speed", StringValue ("Constant:0.0"),
			 "Bounds", RectangleValue (Rectangle (0, 400.0, 0.0, 400.0)));
  mobility.Install (wifiStaNodes01.Get (0));

  positionAllocSta02->Add (Vector (230, 170, 0.0));
  mobility.SetPositionAllocator (positionAllocSta02);
      
  mobility.SetMobilityModel ("ns3::RandomWalk2dMobilityModel",
    			 "Mode", StringValue ("Time"),
			 "Time", StringValue ("2s"),
//			 "Speed", StringValue ("Constant:0.0"),
			 "Speed", RandomVariableValue (UniformVariable (1.0, 3.0)),
			 "Bounds", RectangleValue (Rectangle (0, 400.0, 0.0, 400.0)));
  mobility.Install (wifiStaNodes01.Get (1));

  positionAllocSta03->Add (Vector (270, 320, 0.0));
  mobility.SetPositionAllocator (positionAllocSta03);
      
  mobility.SetMobilityModel ("ns3::RandomWalk2dMobilityModel",
    			 "Mode", StringValue ("Time"),
			 "Time", StringValue ("2s"),
			 "Speed", StringValue ("Constant:0.0"),
			 "Bounds", RectangleValue (Rectangle (0, 400.0, 0.0, 400.0)));
  mobility.Install (wifiStaNodes02.Get (0));

  positionAllocSta04->Add (Vector (330, 280, 0.0));
  mobility.SetPositionAllocator (positionAllocSta04);
      
  mobility.SetMobilityModel ("ns3::RandomWalk2dMobilityModel",
    			 "Mode", StringValue ("Time"),
			 "Time", StringValue ("2s"),
			 "Speed", StringValue ("Constant:0.0"),
			 "Bounds", RectangleValue (Rectangle (0, 400.0, 0.0, 400.0)));
  mobility.Install (wifiStaNodes02.Get (1));

//  NodeContainer adhocNode;
//  adhocNode.Create (1);
  NodeContainer adhoc (wifiStaNodes01, wifiStaNodes02);
  WifiHelper wifiAdhoc;
  NqosWifiMacHelper macAdhoc = NqosWifiMacHelper::Default ();
  macAdhoc.SetType ("ns3::AdhocWifiMac");
  wifiAdhoc.SetRemoteStationManager ("ns3::ConstantRateWifiManager",
                                "DataMode", StringValue ("OfdmRate6Mbps"));
  YansWifiPhyHelper wifiAdhocPhy = YansWifiPhyHelper::Default ();
  YansWifiChannelHelper wifiAdhocChannel = YansWifiChannelHelper::Default ();
  wifiAdhocPhy.SetChannel (wifiAdhocChannel.Create ());
  NetDeviceContainer adhocDevices = wifiAdhoc.Install (wifiAdhocPhy, macAdhoc, adhoc);
/*
  Ptr<ListPositionAllocator> positionAllocSta05 = CreateObject<ListPositionAllocator> ();

  positionAllocSta05->Add (Vector (65, 150, 0.0));
  mobility.SetPositionAllocator (positionAllocSta05);
  // setup the AP.
  mobility.SetMobilityModel ("ns3::RandomWalk2dMobilityModel",
  			 "Mode", StringValue ("Time"),
			 "Time", StringValue ("2s"),
			 "Speed", RandomVariableValue (UniformVariable (1.0, 3.0)),
//			 "Speed", StringValue ("Constant:0.0"),
			 "Bounds", RectangleValue (Rectangle (0, 400.0, 0.0, 400.0)));
  mobility.Install (adhocNode.Get (0));
*/
  InternetStackHelper stack;
  static MeDeHaDtnRoutingHelper mdrouting;
  stack.Install (wifiApNode);

  static Ipv4GlobalRoutingHelper globalRouting;
  static Ipv4ListRoutingHelper listRouting;
  listRouting.Add (mdrouting, 4);
  listRouting.Add (globalRouting, -10);
  stack.SetRoutingHelper (listRouting);

//  stack.SetRoutingHelper (mdrouting);
//  stack.Install (adhoc);
  stack.Install (wifiStaNodes01);
  stack.Install (wifiStaNodes02);

  Ipv4AddressHelper address;

  address.SetBase ("10.1.3.0", "255.255.255.0");
  Ipv4InterfaceContainer p2pInterfaces01, p2pInterfaces02, p2pInterfaces03;
  p2pInterfaces01 = address.Assign (p2pDevices01);

  address.SetBase ("12.1.3.0", "255.255.255.0");
  p2pInterfaces02 = address.Assign (p2pDevices02);

  address.SetBase ("14.1.3.0", "255.255.255.0");
  p2pInterfaces03 = address.Assign (p2pDevices03);

  address.SetBase ("192.168.6.0", "255.255.255.0");
  Ipv4InterfaceContainer adhocInterfaces;
  adhocInterfaces = address.Assign (adhocDevices);

  address.SetBase ("192.168.0.0", "255.255.255.0");
  Ipv4InterfaceContainer staInterfaces01 = address.Assign (staDevices01);
  Ipv4InterfaceContainer staInterfaces02 = address.Assign (staDevices02);
  Ipv4InterfaceContainer apInterfaces = address.Assign (apDevices);

  Address dest01, dest02, dest03, dest04, dest05, dest06, dest07, dest08, dest09;
  std::string protocol;

//  dest01 = InetSocketAddress (staInterfaces01.GetAddress (1), 1025);
//  dest02 = InetSocketAddress (staInterfaces02.GetAddress (1), 1035);
//  dest03 = InetSocketAddress (adhocInterfaces.GetAddress (1), 1045);
  dest03 = InetSocketAddress (staInterfaces01.GetAddress (1), 1045);
  protocol = "ns3::UdpSocketFactory";
  
  OnOffHelper onoff01 = OnOffHelper (protocol, dest03);
  onoff01.SetAttribute ("OnTime", RandomVariableValue (ConstantVariable (1)));
  onoff01.SetAttribute ("OffTime", RandomVariableValue (ConstantVariable (0)));
  onoff01.SetAttribute ("DataRate", DataRateValue (DataRate ("1kbps")));
  ApplicationContainer apps00 = onoff01.Install (wifiStaNodes02.Get (0));
  apps00.Start (Seconds (20.0));
  apps00.Stop (Seconds (1400.0));

  // Create a packet sink to receive these packets
  PacketSinkHelper sink01 ("ns3::UdpSocketFactory",
                         InetSocketAddress (Ipv4Address::GetAny (), 1045));

//  ApplicationContainer apps01 = sink01.Install (adhoc.Get (1));
  ApplicationContainer apps01 = sink01.Install (wifiStaNodes01.Get (1));
  apps01.Start (Seconds (2.0));
  apps01.Stop (Seconds (1680.0));
/*
  OnOffHelper onoff02 = OnOffHelper (protocol, dest02);
  onoff02.SetAttribute ("OnTime", RandomVariableValue (ConstantVariable (1)));
  onoff02.SetAttribute ("OffTime", RandomVariableValue (ConstantVariable (0)));
  onoff02.SetAttribute ("DataRate", DataRateValue (DataRate ("1kbps")));
  ApplicationContainer apps10 = onoff02.Install (wifiStaNodes02.Get (0));
  apps10.Start (Seconds (2.5));
  apps10.Stop (Seconds (1400.0));

  // Create a packet sink to receive these packets
  PacketSinkHelper sink02 ("ns3::UdpSocketFactory",
                         InetSocketAddress (Ipv4Address::GetAny (), 1035));

  ApplicationContainer apps11 = sink02.Install (wifiStaNodes02.Get (1));
  apps11.Start (Seconds (1.0));
  apps11.Stop (Seconds (1480.0));
*/
  Ipv4GlobalRoutingHelper::PopulateRoutingTables ();
  phy.EnablePcapAll ("test-medeha");

  std::cout << "Src 01 Address = " << staInterfaces02.GetAddress (0) << "\n";
  std::cout << "Dst 01 Address = " << adhocInterfaces.GetAddress (1) << "\n";
//  std::cout << "Sta 01 Address = " << staInterfaces02.GetAddress (0) << "\n";
//  std::cout << "Sta 02 Address = " << staInterfaces02.GetAddress (1) << "\n";
  std::cout << "P2P 01 Address = " << p2pInterfaces01.GetAddress (0) << "\n";
  std::cout << "P2P 02 Address = " << p2pInterfaces01.GetAddress (1) << "\n";
  std::cout << "AP 01 Address = " << apInterfaces.GetAddress (0) << "\n";
  std::cout << "AP 02 Address = " << apInterfaces.GetAddress (1) << "\n";

  std::ofstream os;
  os.open ("test-medeha.mob");
  MobilityHelper::EnableAsciiAll (os);

  Simulator::Stop (Seconds (1800.0));

  Simulator::Run ();
  Simulator::Destroy ();
  return 0;
}
