/* -*-  Mode: C++; c-file-style: "gnu"; indent-tabs-mode:nil; -*- */
/*
 * Copyright (c) 2010 INRIA, Sophia Antipolis
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as 
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Author: Rao Naveed Bin Rais <Naveed-Bin.Rais@inria.fr>
 */

#ifndef BUFFER_H_
#define BUFFER_H_


#include <stdlib.h>
#include <list>
#include "ns3/ipv4-address.h"
#include "ns3/packet.h"

#define  LP_PKT_DROP  00
#define  HP_PKT_DROP  01
#define  TTL_EXP_DROP 10

namespace ns3 {


typedef struct
{
  Ipv4Address	nodeAddress;
  int  lowPriorityLossCounter;
  int  highPriorityLossCounter;
  int  ttlExpiredCounter;
}DropPacketsInfoSt;


typedef struct
{
  Ptr<Packet> packet;
  Ipv4Address src;
  Ipv4Address dst;
  uint16_t seqNum;
  int protocol;
  int ttl;
  int priority;
  int16_t copies;
  bool bbCopyGiven;

}BufferedFramesSt;


class BufferClass
{
  public: 
  BufferClass();	// Constructor

  typedef std::list<BufferedFramesSt> BufferedFramesList;
  BufferedFramesList m_pktList;
  int m_listSize;
// list<struct dropPacketsInfo_t*> ptrDropPacketsInfo_;
// struct dropPacketsInfo_t dropPacketsInfo_[10];
// int sizeDPI_;

    
  inline int GetSizeList() {return m_listSize;}
  inline void SetBufferLimitList(int limit) {m_listBufferLimit = limit;}
    
  void AddElementToList(BufferedFramesSt frameToAdd);
//  void RemoveElementFromList(std::list<Ieee80211DataFrame*>::iterator newIterator);
  int IsPacketTagged(BufferedFramesSt frameToAdd);
  void DecrementTTLValueOfElements(void);
//  void CheckAndAddEntryInDropList(nsaddr_t destAddress, int type);

  void SetBufferLimit (uint32_t size);
  uint32_t GetBufferLimit (void);
  uint32_t GetBufferSize (void);

  int m_lowPriorityLossCounter;
  int m_highPriorityLossCounter;
  int m_ttlExpiredCounter;
    
  int m_listBufferLimit;
  
  enum BufferState
    {
      MEMORY_EMPTY = 1,
      MEMORY_AVAILABLE = 2,
      MEMORY_APPROACHING_CRITICAL = 3,
      MEMORY_CRITICAL = 4,
      MEMORY_FULL = 5,
    };
    
  BufferState m_state;
  
  void SetMemoryState (void);
  BufferState GetMemoryState (void);
};

} // ns3 namespace

#endif
