/* -*-  Mode: C++; c-file-style: "gnu"; indent-tabs-mode:nil; -*- */
/*
 * Copyright (c) 2010 INRIA, Sophia Antipolis
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as 
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Author: Rao Naveed Bin Rais <Naveed-Bin.Rais@inria.fr>
 */

#ifndef HANDLE_NOTIFICATIONS_CLASS_H
#define HANDLE_NOTIFICATIONS_CLASS_H

#include <map>
#include <list>
#include "ns3/header.h"
#include "ns3/mac48-address.h"
#include "ns3/packet.h"
#include "ns3/llc-snap-header.h"
#include "ipv4-interface.h"


#define BROADCAST_PACKET_NOTIF      0x0F
#define NORMAL_PACKET_NOTIF         0x01
#define NOTIFICATION_NOTIF          0x10

#define	ASS_NOTIF                   0x10
#define	DISASS_NOTIF                0x11
#define	DEST_FOUND_NOTIF            0x12
#define REMOVE_FRAMES_NOTIF         0x13
#define UPDATE_FRAMES_NOTIF         0x14
#define STORE_FRAMES_NOTIF          0x15
#define FETCH_FRAMES_NOTIF          0x16
#define NEIGHBORHOOD_INFO_NOTIF	    0x17
#define INDIRECT_ASSOCIATION_NOTIF  0x18
#define STA_ASS_NOTIF               0x19
#define NEIGHBORHOOD_DISASS_NOTIF   0x1A
#define MANET_DISCONNECT_NOTIF      0x20


#define SUCCESS                     1
#define FAILURE                     -1

#define DEST_INFO_ENTRY_MODIFIED    0x20
#define DEST_INFO_ENTRY_ADDED       0x21
#define DEST_INFO_ENTRY_REMOVED     0x22
#define DEST_INFO_ENTRY_FOUND       0x23
#define DEST_INFO_ENTRY_NOT_FOUND   0x24
#define DEST_INFO_LIST_DOESNT_EXIST 0x25
#define DEST_INFO_ALTERNATE_FOUND   0x26

#define DEST_BUFFER_INFO_ENTRY_ADDED        0x30
#define	DEST_BUFFER_INFO_ENTRY_MODIFIED     0x31
#define DEST_BUFFER_INFO_ENTRY_FOUND        0x32
#define DEST_BUFFER_INFO_ENTRY_NOT_FOUND    0x33
#define DEST_BUFFER_INFO_ENTRY_REMOVED      0x34

#define BUFFER_LOCAL_AP_RESPONSIBLE_FOR_BUFFER    0x40
#define BUFFER_REMOTE_AP_RESPONSIBLE_FOR_BUFFER   0x41
#define BUFFER_DEST_ENTRY_DOESNT_EXIST            0x42

#pragma once

namespace ns3 {


class MedehaIntraManetNotifHeader : public Header 
{
public:

  MedehaIntraManetNotifHeader ();
  ~MedehaIntraManetNotifHeader ();

  static TypeId GetTypeId (void);
  virtual TypeId GetInstanceTypeId (void) const;
  virtual void Print (std::ostream &os) const;
  virtual uint32_t GetSerializedSize (void) const;
  virtual void Serialize (Buffer::Iterator start) const;
  virtual uint32_t Deserialize (Buffer::Iterator start);
  
  uint16_t GetType (void) { return m_type; }
  void SetType (uint16_t type) { m_type = type; }
  uint16_t GetProtocol (void) { return m_protocol; }
  void SetProtocol (uint16_t protocol) { m_protocol = protocol; }
  Ipv4Address GetDstIPAddress (void) { return m_dstIPAddress; }
  void SetDstIPAddress (Ipv4Address dstIPAddress) { m_dstIPAddress = dstIPAddress; }

private:
  uint16_t  m_type;
  uint16_t  m_protocol;
  Ipv4Address  m_dstIPAddress;
};



class APNotifL3Header : public Header 
{
public:

  APNotifL3Header ();
  ~APNotifL3Header ();

  static TypeId GetTypeId (void);
  virtual TypeId GetInstanceTypeId (void) const;
  virtual void Print (std::ostream &os) const;
  virtual uint32_t GetSerializedSize (void) const;
  virtual void Serialize (Buffer::Iterator start) const;
  virtual uint32_t Deserialize (Buffer::Iterator start);
  
  int GetMsgID (void) { return m_msgID; }
  void SetMsgID (int msgID) { m_msgID = msgID; }
  int GetNumOfIfces (void) { return m_numOfIfces; }
  void SetNumOfIfces (int numOfIfces) { m_numOfIfces = numOfIfces; }
  Ipv4Address GetSrcIPAddress (void) { return m_srcIPAddress; }
  void SetSrcIPAddress (Ipv4Address srcIPAddress) { m_srcIPAddress = srcIPAddress; }
  void SetIPAddresses (Ipv4Address* addrArray, uint16_t len)
  {
    uint16_t i=0;
    while (i < len)
      {
        m_IPAddress[i] = addrArray[i];
	i++;
      }
  }
  Ipv4Address* GetIPAddresses (void) { return m_IPAddress; }
  uint16_t GetManetConnection (void) { return m_manetConnection; }
  void SetManetConnection (uint16_t manetConnection) { m_manetConnection = manetConnection; }
//  void SetIPAddress01 (Ipv4Address assAPIPAddress) { m_IPAddress01 = assAPIPAddress; }
//  Ipv4Address GetIPAddress02 (void) { return m_IPAddress02; }
//  void SetIPAddress02 (Ipv4Address dstIPAddress) { m_IPAddress02 = dstIPAddress; }
//  Ipv4Address GetIPAddress03 (void) { return m_IPAddress03; }
//  void SetIPAddress03 (Ipv4Address dstIPAddress) { m_IPAddress03 = dstIPAddress; }

private:
  uint32_t  m_msgID;
  uint32_t  m_numOfIfces;
  Ipv4Address  m_srcIPAddress;
  Ipv4Address  m_IPAddress[3];
  uint8_t m_manetConnection;
};

typedef struct
{
  Ipv4Address destNodeAddress;
  Ipv4Address respAPAddress;
  Ipv4Address destNodeAltAddress01;
  Ipv4Address destNodeAltAddress02;
}DestBufferInfoSt;

typedef struct
{
  Ipv4Address destNodeAddress;
  Ipv4Address assAPAddress;
  Ipv4Address destNodeAltAddress01;
  Ipv4Address destNodeAltAddress02;
  uint16_t m_manetConnection;  
}DestInfoSt;

class HandleNotificationsClass
{
public:
  HandleNotificationsClass (void);
//  int CheckNotificationMsg (APNotifL3Header* apNotif, Mac48Address myAddress, int *numFrames);
//  void SendNotificationToAPs (int notifType, int numFramesRemoved, Ipv4Address srcIPAddress, Ipv4Address assAPIPAddress, Ipv4Address dstIPAddress, Ptr<Packet> packet);
  void SendNotificationToAPs(int notifType, int numOfIfces, Ipv4Address srcIPAddress, Ipv4Address* ipAddressArray, uint8_t manetConnection, Ptr<Packet> packet);
  int ParseNotification (Ptr<Packet> packet, APNotifL3Header apNotif);
  int CheckForDestInfo (Ipv4Address destAddres, Ipv4Address* assAPAddress, Ipv4Address* infraStructureAddress);
  int AddOrModifyDestInfo(Ipv4Address* addrArray, Ipv4Address assAPIPAddress, uint16_t len, uint8_t manetConnection);
//  int AddOrModifyDestInfo (Ipv4Address srcIPAddress, Ipv4Address assAPIPAddress, Ipv4Address alternateAddress);
//  int RemoveDestInfo (Ipv4Address srcIPAddress, Ipv4Address *alternateAddress);
  int RemoveDestInfo(Ipv4Address srcIPAddress, Ipv4Address *alternateAddress, uint16_t* len);
//  int CheckAndAddDestBufferAP (Ipv4Address srcIpAddress, Ipv4Address assAPIpAddress, Ipv4Address alternateAddress);
  int CheckAndAddDestBufferAP (Ipv4Address* addrArray, Ipv4Address assAPIpAddress, uint16_t len);
  int CheckForDestBufferInfo (Ipv4Address destAddress, Ipv4Address* respApAddress);
  int NeedToBufferPacket (Ipv4Address destAddress, std::vector<Ptr<Ipv4Interface> > m_interfaces);
  int RemoveDestBufferInfo(Ipv4Address srcIPAddress);

  typedef std::list<DestInfoSt> DestInfo;
  DestInfo destInfoList;
  typedef std::list<DestBufferInfoSt> DestBufferInfo;
  DestBufferInfo destBufferInfoList;

public:
  ~HandleNotificationsClass(void);
};


} // ns3 namespace

#endif
