/* -*-  Mode: C++; c-file-style: "gnu"; indent-tabs-mode:nil; -*- */
/*
 * Copyright (c) 2010 INRIA, Sophia Antipolis
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as 
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Author: Rao Naveed Bin Rais <Naveed-Bin.Rais@inria.fr>
 */

#define NS_LOG_APPEND_CONTEXT                                   \
  if (GetObject<Node> ()) { std::clog << "[node " << GetObject<Node> ()->GetId () << "] "; }
  
///
/// \brief Gets the delay between a given time and the current time.
///
/// If given time is previous to the current one, then this macro returns
/// a number close to 0. This is used for scheduling events at a certain moment.
///
#define DELAY(time) (((time) < (Simulator::Now ())) ? Seconds (0.000001) : \
                     (time - Simulator::Now () + Seconds (0.000001)))


/// Maximum allowed jitter.
#define MEDEHA_MAXJITTER		(m_helloInterval.GetSeconds () / 4)
#define JITTER (Seconds (UniformVariable().GetValue (0, MEDEHA_MAXJITTER)))

#define ENTRY_EXPIRE_TIME    Seconds (4)

#define  BUFFER
#define  DEBUG
//#define  MANET
#define  RELAY_METRIC_USED
//#define  ENCOUNTER_AND_SOCIAL
//#define  SOCIAL_AFFILIATION
//#define SOCIAL_AFFILIATION_MANET

#define  MEDEHA_MAX_NUM_FRAMES_TO_SEND  10
#define  MAX_NUM_OF_COPIES   10
#define  MEDEHA_MANET_PH_2

#include "medeha-dtn-routing.h"
#include "ns3/socket-factory.h"
#include "ns3/udp-socket-factory.h"
#include "ns3/simulator.h"

#include "ns3/random-variable.h"
#include "ns3/inet-socket-address.h"
//#include "ns3/ipv4-routing-protocol.h"
#include "ns3/ipv4-route.h"
#include "ns3/boolean.h"
#include "ns3/uinteger.h"
#include "ns3/enum.h"
#include "ns3/trace-source-accessor.h"
#include "ns3/ipv4-header.h"
#include "ns3/address-utils.h"

#include "ns3/arp-l3-protocol.h"
#include "ns3/config.h"
//#include "ns3/simulator.h"
#include "ns3/medeha-handle-notification-ip.h"
#include "ns3/seq-header.h"
#include "ns3/udp-header.h"
#include "ns3/ipv4-list-routing.h"
//#include "ns3/medeha-intra-manet-notif-header.h"


NS_LOG_COMPONENT_DEFINE ("MeDeHaDtnRouting");


#define DTN_ROUTING_PORT_NUMBER  600

namespace ns3 {

//using namespace olsr;


class MeDeHaDTNPacketHeader;

MeDeHaDTNPacketHeader::MeDeHaDTNPacketHeader ()
  : m_msgType (0)
{}

MeDeHaDTNPacketHeader::~MeDeHaDTNPacketHeader ()
{}

TypeId 
MeDeHaDTNPacketHeader::GetTypeId (void)
{
  static TypeId tid = TypeId ("ns3::MeDeHaDTNPacketHeader")
    .SetParent<Header> ()
    .AddConstructor<MeDeHaDTNPacketHeader> ()
    ;
  return tid;
}

TypeId 
MeDeHaDTNPacketHeader::GetInstanceTypeId (void) const
{
  return GetTypeId ();
}

void 
MeDeHaDTNPacketHeader::Print (std::ostream &os) const
{
  switch (m_msgType)
    {
    case HELLO_MESSAGE:
      break;
    }
}

uint32_t 
MeDeHaDTNPacketHeader::GetSerializedSize (void) const
{
//  return 12;
  return GetSize ();
}

uint32_t
MeDeHaDTNPacketHeader::GetSize (void) const
{
  uint32_t size = 0;
  if (m_numOfAddresses == 1)
    {
      size = 14;
    }
  else if (m_numOfAddresses == 2)
    {
      size = 18;
    }
  return size;
}

void 
MeDeHaDTNPacketHeader::Serialize (Buffer::Iterator i) const
{
  i.WriteU16 (m_msgType);
  i.WriteU16 (m_numOfAddresses);
  i.WriteU16 (m_associated);
  WriteTo (i, m_associatedWith);
  WriteTo (i, m_IPAddress01);
  if (m_numOfAddresses > 1)
    {
      WriteTo (i, m_IPAddress02);
    }
}

uint32_t
MeDeHaDTNPacketHeader::Deserialize (Buffer::Iterator start)
{
  Buffer::Iterator i = start;
  m_msgType = i.ReadU16 ();
  m_numOfAddresses = i.ReadU16 ();
  m_associated = i.ReadU16 ();
  ReadFrom (i, m_associatedWith);
  ReadFrom (i, m_IPAddress01);
  if (m_numOfAddresses > 1)
    {
      ReadFrom (i, m_IPAddress02);
    }
    
  return i.GetDistanceFrom (start);
}

NS_OBJECT_ENSURE_REGISTERED (MeDeHaDtnRouting);

TypeId 
MeDeHaDtnRouting::GetTypeId (void)
{
  static TypeId tid = TypeId ("ns3::MeDeHaDtnRouting")
    .SetParent<Ipv4RoutingProtocol> ()
    .AddConstructor<MeDeHaDtnRouting> ()
    .AddAttribute ("HelloInterval", "HELLO messages emission interval.",
                   TimeValue (Seconds (2.0)),
                   MakeTimeAccessor (&MeDeHaDtnRouting::m_helloInterval),
                   MakeTimeChecker ())
    ;
  return tid;
}

//class MeDeHaDtnRouting;

MeDeHaDtnRouting::MeDeHaDtnRouting ()
  : m_ipv4 (0),
    m_helloTimer (Timer::CANCEL_ON_DESTROY),    
    m_associated (false),
    m_manet (false),
    m_adhoc (false)
{
  m_defaultRoute = 0;
  m_niInterval = Seconds (5.0);
  m_txTimer = Seconds (0.05);
  m_txTimerSource = Seconds (0.05);
  m_txTimerMultiCopy = Seconds (0.05);
  
  m_manetInterval = Seconds (4.0);
  m_txTimerManet = Seconds (0.05);
}

MeDeHaDtnRouting::~MeDeHaDtnRouting ()
{
  if (m_txTimerEvent.IsRunning ())
    {
      m_txTimerEvent.Cancel ();
    }
    
  if (m_txTimerSourceEvent.IsRunning ())
    {
      m_txTimerSourceEvent.Cancel ();
    }
    
  if (m_niTimerEvent.IsRunning ())
    {
      m_niTimerEvent.Cancel ();
    }
  
  if (m_manetTimerEvent.IsRunning ())
    {
      m_manetTimerEvent.Cancel ();
    }
  
  if (m_txTimerEventManet.IsRunning ())
    {
      m_txTimerEventManet.Cancel ();
    }
}

void
MeDeHaDtnRouting::SetIpv4 (Ptr<Ipv4> ipv4)
{
  NS_ASSERT (ipv4 != 0);
  NS_ASSERT (m_ipv4 == 0);
  NS_LOG_DEBUG ("Created MeDeHaDtnRouting");
  m_helloTimer.SetFunction (&MeDeHaDtnRouting::HelloTimerExpire, this);

  m_ipv4 = ipv4;
  ConnectToCatchAssocEvent ();
  ConnectToCatchManetEvent ();

  Simulator::ScheduleNow (&MeDeHaDtnRouting::Start, this);
}

void MeDeHaDtnRouting::DoDispose ()
{
  m_ipv4 = 0;

  for (std::map< Ptr<Socket>, Ipv4Address >::iterator iter = m_socketAddresses.begin ();
       iter != m_socketAddresses.end (); iter++)
    {
      iter->first->Close ();
    }
  m_socketAddresses.clear ();

  Ipv4RoutingProtocol::DoDispose ();
}

void MeDeHaDtnRouting::Start ()
{
	m_routingProtocolCode = 0x3F;
  if (m_mainAddress == Ipv4Address ())
    {
      Ipv4Address loopback ("127.0.0.1");
      for (uint32_t i = 0; i < m_ipv4->GetNInterfaces (); i++)
        {
          // Use primary address, if multiple
          Ipv4Address addr = m_ipv4->GetAddress (i, 0).GetLocal ();
          if (addr != loopback)
            {
	      // Here, we strongly assume that the first IP address would be the address of the ad hoc network
	      // So, this assumption should be kept in mind when configuring the network's IP addresses
              m_mainAddress = addr;
              break;
            }
        }
	
      if (m_ipv4->GetNInterfaces () >= 3)
        {
	  m_addressAlternate01 = m_ipv4->GetAddress (2, 0).GetLocal ();;
	}
	
      if (m_ipv4->GetNInterfaces () >= 4)
        {
	  m_addressAlternate02 = m_ipv4->GetAddress (3, 0).GetLocal ();;
	}

      NS_ASSERT (m_mainAddress != Ipv4Address ());
    }

  NS_LOG_DEBUG ("Starting MeDeHa DTN Routing on node " << m_mainAddress);
  std::cout << "Starting MeDeHa DTN Routing on node " << m_mainAddress << std::endl;

  Ipv4Address loopback ("127.0.0.1");
  for (uint32_t i = 0; i < m_ipv4->GetNInterfaces (); i++)
    {
      Ipv4Address addr = m_ipv4->GetAddress (i, 0).GetLocal ();
      if (addr == loopback)
        continue;

      if (addr != m_mainAddress)
        {
          // Create never expiring interface association tuple entries for our
          // own network interfaces, so that GetMainAddress () works to
          // translate the node's own interface addresses into the main address.
//          IfaceAssocTuple tuple;
//          tuple.ifaceAddr = addr;
//          tuple.mainAddr = m_mainAddress;
//          AddIfaceAssocTuple (tuple);
//          NS_ASSERT (GetMainAddress (addr) == m_mainAddress);
        }

      // Create a socket to listen only on this interface
      Ptr<Socket> socket = Socket::CreateSocket (GetObject<Node> (), 
        UdpSocketFactory::GetTypeId()); 
      socket->SetRecvCallback (MakeCallback (&MeDeHaDtnRouting::RecvPacket,  this));
      if (socket->Bind (InetSocketAddress (addr, DTN_ROUTING_PORT_NUMBER)))
        {
          NS_FATAL_ERROR ("Failed to bind() MeDeHa DTN Routing receive socket");
        }
      socket->Connect (InetSocketAddress (Ipv4Address (0xffffffff), DTN_ROUTING_PORT_NUMBER));
      m_socketAddresses[socket] = addr;
    }

//  HelloTimerExpire ();
    Time time = JITTER;
    m_helloTimer.Schedule (time);
    
//  olsrRouting->SetManetCallback (MakeCallback (&MeDeHaDtnRouting::ReceiveCallbackFromManet (this));

  NS_LOG_DEBUG ("MeDeHa DTN Routing on node " << m_mainAddress << " started");
}

void MeDeHaDtnRouting::SetMainInterface (uint32_t interface)
{
  m_mainAddress = m_ipv4->GetAddress (interface, 0).GetLocal ();
}


void
MeDeHaDtnRouting::RecvPacket (Ptr<Socket> socket)
{
  Ptr<Packet> receivedPacket;
  Address sourceAddress;
  receivedPacket = socket->RecvFrom (sourceAddress);

  InetSocketAddress inetSourceAddr = InetSocketAddress::ConvertFrom (sourceAddress);
  Ipv4Address senderIfaceAddr = inetSourceAddr.GetIpv4 ();
  Ipv4Address receiverIfaceAddr = m_socketAddresses[socket];
  NS_ASSERT (receiverIfaceAddr != Ipv4Address ());
  NS_LOG_DEBUG ("DTN node " << m_mainAddress << " received a DTN message from " << senderIfaceAddr << " to " << receiverIfaceAddr);
  
  NS_ASSERT (inetSourceAddr.GetPort () == DTN_ROUTING_PORT_NUMBER);
  
  Ptr<Packet> packet = receivedPacket;

  MeDeHaDtnPeerInfoHeader header;
  packet->RemoveHeader (header);

#ifdef DEBUG
  std::cout << "Type = " << header.GetHeaderType () << " Length = " << header.GetHeaderLength () << "\n";
#endif

  int size = header.GetHeaderLength () - MEDEHA_DTN_PEER_INFO_HEADER_SIZE;
  
  if (header.GetDestAddress ().IsEqual (Ipv4Address (0xFFFFFFFF)) ||
      header.GetDestAddress ().IsEqual (m_mainAddress))
    {
      while (size > 0)
        {
          MeDeHaDtnMessageHeader msg;
          packet->RemoveHeader (msg);
  
          switch (msg.GetMessageType ())
            {      
              case MeDeHaDtnMessageHeader::CURRENT_PEERS:
                ProcessCurrentPeerInfo (msg, header.GetSrcAddress ());
                break;
	 
              case MeDeHaDtnMessageHeader::RECENT_PEERS:
                ProcessRecentPeerInfo (msg, header.GetSrcAddress ());
                break;
	
              case MeDeHaDtnMessageHeader::HELLO_MSG:
                ProcessHello (msg);
                break;
		
              case MeDeHaDtnMessageHeader::MSG_VECTOR:
#ifdef DEBUG
		std::cout << "Processing Message Vector Message\n";
#endif
                ProcessMessageVector (msg);
                break;
		
              case MeDeHaDtnMessageHeader::MANET_PEERS:
                ProcessManetPeerInfo (msg, header.GetSrcAddress ());
                break;
            }
          size -= msg.GetSerializedSize ();
        }
    }  
}


void
MeDeHaDtnRouting::ProcessHello (MeDeHaDtnMessageHeader msg)
{
//  m_associated = msg.GetAssociated ();
//  m_associatedWith = msg.GetAssociatedWith ();

  m_adhoc = true;
    
  Ipv4Address address[3];
 
  MeDeHaDtnMessageHeader::HelloMessage hello = msg.GetHelloMessage ();
  std::vector<Ipv4Address>::iterator iter = hello.m_interfaces.interfaceAddresses.begin ();
  for (int i=0; i<hello.m_interfaces.m_numOfIfces; i++, iter++)
    {
      address[i] = *iter;
    }
    
#ifdef DEBUG
  std::cout << "\nDTN node: " << m_mainAddress << ", HELLO message is received and contains:\n";
  std::cout << "Associated Flag = " << hello.GetAssociated () << ",\nCorresponding AP = " << hello.GetAssociatedWith () << ",\nBuffer State = " << hello.GetBufferState () << ",\nNode Address 01 = " << address[0] << ",\nNode Address 02 = " << address[1] << "\nNode Address 03 = " << address[2] << "\n";
#endif

/*
  if (hello.GetManet () &&
      m_manet)
    {
      return;
    }
*/      
/*
  if (header.GetNumOfAddresses () == 2)
    {
      status = RefreshEntryForDest (header.GetIPAddress02 ());
    }
*/
  if (hello.GetBufferState () != BufferClass::MEMORY_FULL ||
      hello.GetBufferState () != BufferClass::MEMORY_CRITICAL)
    {
      UpdateRoutingTable (address, hello.m_interfaces.m_numOfIfces, address[0], true);
  
      bool status = 0;
      status = RefreshEntryForDest (address[0], address[0], address, hello.m_interfaces.m_numOfIfces, 1);
      UpdateContactTable (address, hello.m_interfaces.m_numOfIfces, status);
      
      if (hello.m_interfaces.m_numOfIfces >= 2)
        {
	  status = RefreshEntryForDest (address[1], address[0], address, hello.m_interfaces.m_numOfIfces, 0);
	}
	
      if (hello.m_interfaces.m_numOfIfces >= 3)
        {
	  status = RefreshEntryForDest (address[2], address[0], address, hello.m_interfaces.m_numOfIfces, 0);
	}
      
      Ptr<Ipv4L3Protocol> ipv4L3 = m_node->GetObject <Ipv4L3Protocol> ();
/*      
      if (m_niTimerEvent.IsRunning ())
	{
	  
	}
      else
	{
*/	  // inform AP about this new change
	  // has to check if state
	  if (m_associated)
	    {          
	      HandleNotificationsClass hNotif = ipv4L3->GetHandleNotificationClass ();
      
              Ptr<Packet> packetToSend = Create<Packet> ();
      
	      Ipv4Address addr;
	      for (uint32_t i = 0; i < m_ipv4->GetNInterfaces (); i++)
		{
		  addr = m_ipv4->GetAddress (i, 0).GetLocal ();
		  Ipv4Address loopback ("127.0.0.1");
	  
		  if (addr != loopback &&
		      addr != m_mainAddress)
		    {
		      // Here, we strongly assume that the second IP addreses in the list of interfaces is the IP address of infrastructure network
		      // So, it should be kept in mind when configuring network's IP addresses
		      m_addressAlternate01 = addr;
		      break;
		    }
		}

	      // header.GetIPAddress01 () signifies the ad hoc interface address of a station
	      // while sending information to an associated AP, we keep the infrastructure address "header.GetIPAddress02 ()" of the station first
              hNotif.SendNotificationToAPs (NEIGHBORHOOD_INFO_NOTIF, hello.m_interfaces.m_numOfIfces, addr, address, 0, packetToSend);
              ipv4L3->Send (packetToSend, m_mainAddress, m_associatedWith, 17, NULL);
	    }
//	  m_niTimerEvent = Simulator::Schedule (m_niInterval, &MeDeHaDtnRouting::HandleNITimer, this);
//	}
    
      int numStoredFrames01 = CountNumberOfStoredFrames (address[0]);
      int numStoredFrames02 = CountNumberOfStoredFrames (address[1]);
      int numStoredFrames03 = CountNumberOfStoredFrames (address[2]);
      
#ifdef DEBUG
      std::cout << "MeDeHaDtnRouting::ProcessHello, Number of Stored Frames for dest = " << address[1] << " are " << numStoredFrames01 << "\n";
#endif
	
      // check if there are still some frames to be sent for the destination "tempAddress"
      if (numStoredFrames01 > 0 ||
          numStoredFrames02 > 0 ||
	  numStoredFrames03 > 0)
	{
	  if (m_txTimerEvent.IsRunning ())
	      m_txTimerEvent.Cancel ();
	  // send frames to the corresponding AP
	  SendStoredFrames (ipv4L3);
	  m_txTimerEvent = Simulator::Schedule (m_txTimer, &MeDeHaDtnRouting::HandleSendTimer, this);	// reschedule the timer for transmission of buffered frames
	}

      // Again, the assumption is that address[0] always contains IP address of the ad hoc interface
      SendPeerInfo (address[0]);
    }
}

void
MeDeHaDtnRouting::ProcessCurrentPeerInfo (MeDeHaDtnMessageHeader msg, Ipv4Address senderAddress)
{
  MeDeHaDtnMessageHeader::CurrentPeerInfo currentPI = msg.GetCurrentPeerInfo ();
  
#ifdef DEBUG
  std::cout << "\nDTN node: " << m_mainAddress << ", CURRENT_PEER_INFO message is received from " << senderAddress << " and contains:\n";
  std::cout << "Number of Stations = " << (int) msg.GetNumOfStations () << "\n";
#endif

  std::vector<MeDeHaDtnMessageHeader::Interfaces>::iterator iterIfces = currentPI.m_interfaces.begin ();
  
//  std::cout << "sender address = " << senderAddress << std::endl;
  RemoveRouteWithGWAsDest (senderAddress);
  for (int i=0; i<msg.GetNumOfStations (); i++, iterIfces++)
    {
      Ipv4Address address[3];
      std::vector<Ipv4Address>::iterator iter = iterIfces->interfaceAddresses.begin ();
      for (int j=0; j<iterIfces->m_numOfIfces; j++, iter++)
        {	  
          address[j] = *iter;
        }
	
      if (address[0] == m_mainAddress ||
          address[1] == m_mainAddress ||
          address[2] == m_mainAddress)
        {
          continue;
        }

#ifdef DEBUG
      std::cout << "address1: " << address[0] << " address2: " << address[1] << " address3: " << address[2] << "\n";
#endif

      RefreshEntryForDest (address[0], senderAddress, address, iterIfces->m_numOfIfces, 0);
      UpdateRoutingTable (address, iterIfces->m_numOfIfces, senderAddress, 0);
      
      if (iterIfces->m_numOfIfces >= 2)
        {
          RefreshEntryForDest (address[1], senderAddress, address, iterIfces->m_numOfIfces, 0);
        }

      if (iterIfces->m_numOfIfces >= 3)
        {
          RefreshEntryForDest (address[2], senderAddress, address, iterIfces->m_numOfIfces, 0);
        }

      Ptr<Ipv4L3Protocol> ipv4L3 = m_node->GetObject <Ipv4L3Protocol> ();

      int numStoredFrames01 = CountNumberOfStoredFrames (address[0]);
      int numStoredFrames02 = CountNumberOfStoredFrames (address[1]);
      int numStoredFrames03 = CountNumberOfStoredFrames (address[2]);
      
#ifdef DEBUG
      std::cout << "MeDeHaDtnRouting::ProcessCurrentPeerInfo, Number of Stored Frames for dest = " << address[1] << " are " << numStoredFrames01 << "\n";
#endif

      // check if there are still some frames to be sent for the destination "tempAddress"
      if (numStoredFrames01 > 0 ||
          numStoredFrames02 > 0 ||
          numStoredFrames03 > 0)
        {
          if (m_txTimerEvent.IsRunning ())
              m_txTimerEvent.Cancel ();
          // send frames to the corresponding AP
          SendStoredFrames (ipv4L3);
          m_txTimerEvent = Simulator::Schedule (m_txTimer, &MeDeHaDtnRouting::HandleSendTimer, this);	// reschedule the timer for transmission of buffered frames
        }

#ifdef SOCIAL_AFFILIATION_MANET
      if (buffer.GetBufferSize ())
        {
    	  Ipv4Address manetAddress = GetManetAddressFromAdhoc (senderAddress);
    	  Ipv4Address myManetAddress = GetManetAddressFromAdhoc (m_mainAddress);
    	  uint32_t peer = (manetAddress.Get () >> 8) & 0xff;
    	  uint32_t mine = (myManetAddress.Get () >> 8) & 0xff;
//	  std::cout << "peer = " << manetAddress << ", mine = " << myManetAddress << " " << peer << " " << mine << std::endl;
	  
    	  if (manetAddress != "102.102.102.102" &&
              peer != mine)
    	    {
              ForwardStoredFramesToManetSocialNode (ipv4L3, manetAddress, senderAddress);
            }
        }
#endif

#ifdef SOCIAL_AFFILIATION
#ifdef DEBUG
      std::cout << "Buffer Size at " << m_mainAddress << ": " << buffer.GetBufferSize () << std::endl;
#endif
      if (buffer.GetBufferSize ())
        {
          ForwardStoredFramesToSocialNode (ipv4L3, senderAddress);
        }
#endif
    }
}

void
MeDeHaDtnRouting::ProcessRecentPeerInfo (MeDeHaDtnMessageHeader msg, Ipv4Address senderAddress)
{
  MeDeHaDtnMessageHeader::RecentPeerInfo recentPI = msg.GetRecentPeerInfo ();

#ifdef DEBUG  
  std::cout << "\nDTN node: " << m_mainAddress << ", RECENT_PEER_INFO message is received from " << senderAddress << " and contains:\n";
  std::cout << "Number of Stations = " << (int) msg.GetNumOfStations () << "\n\n";
#endif

  std::vector<MeDeHaDtnMessageHeader::StaList>::iterator iterStaList = recentPI.m_staList.begin ();
  
  for (int i=0; i<msg.GetNumOfStations (); i++, iterStaList++)
    {
      Ipv4Address address[3];
      std::vector<Ipv4Address>::iterator iter = iterStaList->interfaceAddresses.begin ();
      for (int j=0; j<iterStaList->m_numOfIfces; j++, iter++)
        {
          address[j] = *iter;
        }
//      UpdateRoutingTable (address[0], address[1], senderAddress, 0);
//      RefreshEntryForDest (address [0], senderAddress, 0);
      
      std::vector<MeDeHaDtnMessageHeader::Metric>::iterator iter2 = iterStaList->m_metricList.begin ();
      int numEncounters = 0;
      Time encounterTime;
      
      for (int j=0; j<iterStaList->m_numOfMetrics; j++, iter2++)
        {
          int type = iter2->m_metricType;
          float value = iter2->m_metricValue;
	  
#ifdef DEBUG
          std::cout << address[0] << ": type = " << type << "\t" << "value = " << value << "\n";
#endif

          if (type == METRIC_ENCOUNTER_NUMBERS)
            {
              numEncounters = (int) value;
            }
	
          else if (type == METRIC_ENCOUNTER_TIME)
            {
              encounterTime = Seconds (value);
            }
        }

#ifdef RELAY_METRIC_USED	      
      int numStoredFrames = CountNumberOfStoredFrames (address[0]);
      int numStoredFrames01 = CountNumberOfStoredFrames (address[1]);
      int numStoredFrames02 = CountNumberOfStoredFrames (address[2]);
      
      if (numStoredFrames > 0 ||
          numStoredFrames01 > 0 ||
          numStoredFrames02 > 0)
        {
          Ptr<Ipv4L3Protocol> ipv4L3 = m_node->GetObject <Ipv4L3Protocol> ();      
          MeDeHaDtnRouting::ContactTable* ct = SearchForDestEntryInContactTable (address[0]);

//          bool flag = CompareSocialAffiliation (address[0], senderAddress);
#ifdef ENCOUNTER_AND_SOCIAL
       	  bool flag = false;
	  Ipv4Address manetAddress = GetManetAddressFromAdhoc (senderAddress);
    	  Ipv4Address myManetAddress = GetManetAddressFromAdhoc (m_mainAddress);
    	  uint32_t peer = (manetAddress.Get () >> 8) & 0xff;
    	  uint32_t mine = (myManetAddress.Get () >> 8) & 0xff;
	  
	  if (peer != mine)
	    {
	      flag = CompareSocialAffiliationForManet (address[2], manetAddress);
	    }
#endif
          
	  if (ct != NULL)
            {
#ifndef ENCOUNTER_AND_SOCIAL
              if ((ct->encounterTime < encounterTime ||
                   ct->numOfEncounters < numEncounters) &&
                   numEncounters >= 2)
#endif

#ifdef ENCOUNTER_AND_SOCIAL
              if ((ct->encounterTime < encounterTime ||
                   ct->numOfEncounters < numEncounters) &&
                   numEncounters >= 2 &&
		   flag == true)
#endif
                {
                  // Add route to this destination with senderAddress as next hop
                  // Send Stored Frames for address[0] to senderAddress
                  RefreshEntryForDest (address[0], senderAddress, address, iterStaList->m_numOfIfces, 0);
                  UpdateRoutingTable (address, iterStaList->m_numOfIfces, senderAddress, 0);
	      
                  if (iterStaList->m_numOfIfces >= 2)
                    {
                      RefreshEntryForDest (address[1], senderAddress, address, iterStaList->m_numOfIfces, 0);
                    }
		    
                  if (iterStaList->m_numOfIfces >= 3)
                    {
                      RefreshEntryForDest (address[2], senderAddress, address, iterStaList->m_numOfIfces, 0);
                    }

//#ifdef DEBUG
                  std::cout << "Forwarding messages to relay 01 " << senderAddress << "\n";
//#endif
                  m_destMultiCopy01 = address[0];
                  m_destMultiCopy02 = address[1];
                  m_destMultiCopy03 = address[2];
                  CompareSeqNumAndSendPacket (ipv4L3, m_destMultiCopy01, m_destMultiCopy02, m_destMultiCopy03);
//                  CompareSeqNumAndSendPacket (ipv4L3, address [1]);
                }
            }
          else
            {
#ifndef ENCOUNTER_AND_SOCIAL
              if (numEncounters >= 2)
#endif

#ifdef ENCOUNTER_AND_SOCIAL
              if (numEncounters >= 2 &&
                  flag == true)
#endif

                {
                  RefreshEntryForDest (address[0], senderAddress, address, iterStaList->m_numOfIfces, 0);
                  UpdateRoutingTable (address, iterStaList->m_numOfIfces, senderAddress, 0);
	      
                  if (iterStaList->m_numOfIfces >= 2)
                    {
                      RefreshEntryForDest (address[1], senderAddress, address, iterStaList->m_numOfIfces, 0);
                    }
		    
                  if (iterStaList->m_numOfIfces >= 3)
                    {
                      RefreshEntryForDest (address[2], senderAddress, address, iterStaList->m_numOfIfces, 0);
                    }
		    	      
//#ifdef DEBUG
                  std::cout << "Forwarding messages to relay 02 " << senderAddress << "\n";
//#endif
                  m_destMultiCopy01 = address[0];
                  m_destMultiCopy02 = address[1];
                  m_destMultiCopy03 = address[2];
                  CompareSeqNumAndSendPacket (ipv4L3, m_destMultiCopy01, m_destMultiCopy02, m_destMultiCopy03);
//                  CompareSeqNumAndSendPacket (ipv4L3, address [1]);
                }
            }
        }
#endif
    }
}


void
MeDeHaDtnRouting::ProcessMessageVector (MeDeHaDtnMessageHeader msg)
{
  MeDeHaDtnMessageHeader::MessageVectorList msgVector = msg.GetMessageVectorList ();
  
#ifdef DEBUG
  std::cout << "\nDTN node: " << m_mainAddress << ", MSG_VECTOR message is received and contains:\n";
  std::cout << "Number of Messages = " << msgVector.GetNumOfMsgs () << "\n";
#endif

  m_msgVectorSummeryList.clear ();
  std::vector<MeDeHaDtnMessageHeader::MessageVector>::iterator iterMsgVector = msgVector.m_msgVectorList.begin ();
  
  for (int i=0; i<msgVector.GetNumOfMsgs (); i++)
    {
      VectorSummery msgVectorSummery;
      msgVectorSummery.srcAddress = iterMsgVector->srcAddress;
      msgVectorSummery.dstAddress = iterMsgVector->dstAddress;
      msgVectorSummery.seqNumber = iterMsgVector->seqNumber;
      m_msgVectorSummeryList.push_back (msgVectorSummery);
    }  
}


void
MeDeHaDtnRouting::SendPacket (Ptr<Packet> packet)
{
  NS_LOG_DEBUG ("DTN Routing node " << m_mainAddress << " sending a DTN Routing packet");

  // Send it
  m_socketAddresses.begin ()->first->Send (packet);
}


void
MeDeHaDtnRouting::SendHello ()
{
  NS_LOG_FUNCTION (this);
  Ptr<Packet> packet = Create<Packet> ();
  
  uint16_t addressCounter = 0;
//  Ipv4Address ipv4AddressAlternate;
//  uint16_t interfaceId[2];
  Ipv4Address address[3];
  
  Ipv4Address loopback ("127.0.0.1");
  for (uint32_t i = 0; i < m_ipv4->GetNInterfaces (); i++)
    {
      Ipv4Address addr = m_ipv4->GetAddress (i, 0).GetLocal ();
      if (addr != loopback &&
          addr != m_mainAddress)
        {
//          ipv4AddressAlternate = addr;
          address[addressCounter] = addr;
          addressCounter++;
        }
    }
 
  MeDeHaDtnMessageHeader msg;

  msg.SetNumOfStations (1);
  
  MeDeHaDtnMessageHeader::HelloMessage &helloMsg = msg.GetHelloMessage ();

  helloMsg.SetAssociated (m_associated);
  helloMsg.SetAssociatedWith (m_associatedWith);
  helloMsg.SetManet (m_manet);
  helloMsg.SetBufferState (buffer.GetMemoryState ());
  
  helloMsg.m_interfaces.m_numOfIfces = addressCounter+1;
  helloMsg.m_interfaces.interfaceAddresses.push_back (m_mainAddress);

  uint16_t j=0;
  while (j < addressCounter)
    {
      helloMsg.m_interfaces.interfaceAddresses.push_back (address[j]);
      j++;
    }

  int size = msg.GetSerializedSize ();  
  packet->AddHeader (msg);
  
  MeDeHaDtnPeerInfoHeader header;
  
  header.SetHeaderType (MEDEHA_DTN_HEADER);
  header.SetHeaderLength (size + MEDEHA_DTN_PEER_INFO_HEADER_SIZE);
  header.SetDestAddress (Ipv4Address (0xFFFFFFFF));
  header.SetSrcAddress (m_mainAddress);
  
  packet->AddHeader (header);
  
//  std::cout << "Having " << addressCounter << " addresses, Sending MeDeHa Header Type = " << header.GetHeaderType () << "\n";
  
  SendPacket (packet);
}

void
MeDeHaDtnRouting::SendPeerInfo (Ipv4Address destAddress)
{
  Ptr<Packet> packet = Create<Packet> ();
  
  MeDeHaDtnMessageHeader msg01, msg02, msg03, msg04;
  int numOfStations = 0;
  int size = 0;

  numOfStations = 0;
  MeDeHaDtnMessageHeader::RecentPeerInfo &recentPI = msg02.GetRecentPeerInfo ();
  
  CTbleI iter;
  
  for (iter = m_contactTable.begin (); iter != m_contactTable.end (); iter++)
    {
      if (iter->staAddress == destAddress ||
          iter->staAltAddress01 == destAddress ||
	  iter->staAltAddress02 == destAddress)
        {
          continue;
        }
	
      bool flag = false;
      flag = SearchForDestEntryInRoutingTable (iter->staAddress);
      
      if (flag)
        {
          continue;
        }
      
      flag = false;
      Ipv4Address addresses[3];
      addresses[0] = iter->staAddress;
      addresses[1] = iter->staAltAddress01;
      addresses[2] = iter->staAltAddress02;
      flag = SearchForDestEntryInManetRoutingTable (addresses, 3);
      
      if (flag)
        {
	  continue;
	}

      MeDeHaDtnMessageHeader::StaList staList;
      staList.interfaceAddresses.push_back (iter->staAddress);
      
      if (iter->staAltAddress01.IsEqual ("102.102.102.102"))
        {
          staList.m_numOfIfces = 1; 
        }
      else if (iter->staAltAddress02.IsEqual ("102.102.102.102"))
        {
          staList.interfaceAddresses.push_back (iter->staAltAddress01);
          staList.m_numOfIfces = 2;
        }
      else
        {
          staList.interfaceAddresses.push_back (iter->staAltAddress01);
          staList.interfaceAddresses.push_back (iter->staAltAddress02);
          staList.m_numOfIfces = 3;
        }
	
      MeDeHaDtnMessageHeader::Metric metric;
      metric.m_metricType = METRIC_ENCOUNTER_NUMBERS;
      metric.m_metricValue = iter->numOfEncounters;
      staList.m_metricList.push_back (metric);
      
      metric.m_metricType = METRIC_ENCOUNTER_TIME;
      metric.m_metricValue = iter->encounterTime.GetSeconds ();
      staList.m_metricList.push_back (metric);

      staList.m_numOfMetrics = 2;
      
      recentPI.m_staList.push_back (staList);
      numOfStations++;
    }
    
  msg02.SetNumOfStations (numOfStations);
  
  size += msg02.GetSerializedSize ();
  packet->AddHeader (msg02);
  
#ifdef MANET
  if (m_manet)
    {
      MeDeHaDtnMessageHeader::ManetPeerInfo &manetPI = msg04.GetManetPeerInfo ();
  
      numOfStations = 0;
      std::vector<olsr::RoutingTableEntry>::iterator it;
      for (it = m_olsrRtEntries.begin (); it != m_olsrRtEntries.end (); it++)
        {
          if (it->destAddr == destAddress)
            {
              continue;
            }
	    
          MeDeHaDtnMessageHeader::Interfaces interfaces;
          interfaces.interfaceAddresses.push_back (it->destAddr);
          interfaces.m_numOfIfces = 1;
      
          manetPI.m_interfaces.push_back (interfaces);
          numOfStations++;
        }
    
      msg04.SetNumOfStations (numOfStations);
  
      size += msg04.GetSerializedSize ();
      packet->AddHeader (msg04);
    }
#endif
    
  MeDeHaDtnMessageHeader::CurrentPeerInfo &currentPI = msg01.GetCurrentPeerInfo ();
  
  numOfStations = 0;
  RTbleI iterator;
   
  for (iterator = m_routingTable.begin (); iterator != m_routingTable.end (); iterator++)
    {
      if ((iterator->staAddress == destAddress &&
           iterator->immediateNeighbor) ||
          (iterator->staAltAddress01 == destAddress &&
           iterator->immediateNeighbor) ||
	  (iterator->staAltAddress02 == destAddress &&
           iterator->immediateNeighbor) || 
          iterator->nextHop == destAddress ||
          !iterator->immediateNeighbor)
        {
          continue;
        }
      MeDeHaDtnMessageHeader::Interfaces interfaces;
      interfaces.interfaceAddresses.push_back (iterator->staAddress);
      
      if (iterator->staAltAddress01.IsEqual ("102.102.102.102"))
        {
          interfaces.m_numOfIfces = 1; 
        }
      else if (iterator->staAltAddress02.IsEqual ("102.102.102.102"))
        {
          interfaces.interfaceAddresses.push_back (iterator->staAltAddress01);
          interfaces.m_numOfIfces = 2;
        }
      else
        {
          interfaces.interfaceAddresses.push_back (iterator->staAltAddress01);
          interfaces.interfaceAddresses.push_back (iterator->staAltAddress02);
          interfaces.m_numOfIfces = 3;
        }
      
      currentPI.m_interfaces.push_back (interfaces);
      numOfStations++;
    }
    
  msg01.SetNumOfStations (numOfStations);
  
  size += msg01.GetSerializedSize ();
  packet->AddHeader (msg01);
  
  int bufferSize = buffer.GetBufferSize ();
  numOfStations = 0;

  if (bufferSize > 0)
    {  
      MeDeHaDtnMessageHeader::MessageVectorList &msgVector = msg03.GetMessageVectorList ();  

      msgVector.SetNumOfMsgs (bufferSize); 
/*
      VSummeryI msgIter;
      for (msgIter = m_msgVectorSummeryList.begin (); msgIter != m_msgVectorSummeryList.end (); msgIter++)
        {
          MeDeHaDtnMessageHeader::MessageVector vecSum;
          vecSum.srcAddress = msgIter->srcAddress;
          vecSum.dstAddress = msgIter->dstAddress;
          vecSum.seqNumber = msgIter->seqNumber;
          msgVector.m_msgVectorList.push_back (vecSum);
        }
*/
      msgVector.m_msgVectorList.clear ();
      std::list<BufferedFramesSt>::iterator newIterator;		// link list iterator
      for (newIterator = buffer.m_pktList.begin(); newIterator != buffer.m_pktList.end(); newIterator++)
        {
          MeDeHaDtnMessageHeader::MessageVector vecSum;
          vecSum.srcAddress = newIterator->src;
          vecSum.dstAddress = newIterator->dst;
          vecSum.seqNumber = newIterator->seqNum;
          msgVector.m_msgVectorList.push_back (vecSum);
        }
		  	
      msg03.SetNumOfStations (numOfStations);
      
      size += msg03.GetSerializedSize ();
      packet->AddHeader (msg03);
    }  
  
  MeDeHaDtnPeerInfoHeader header;
  
  header.SetHeaderType (MEDEHA_DTN_HEADER);
  header.SetHeaderLength (size + MEDEHA_DTN_PEER_INFO_HEADER_SIZE);
  header.SetDestAddress (Ipv4Address (0xFFFFFFFF));
  header.SetSrcAddress (m_mainAddress);

//  std::cout << "size is " << size << "\n";
      
  packet->AddHeader (header);

  SendPacket (packet);
}

///
/// \brief Sends a HELLO message and reschedules the HELLO timer.
/// \param e The event which has expired.
///
void
MeDeHaDtnRouting::HelloTimerExpire ()
{
  if (Simulator::Now ().GetSeconds () < 7198)
    {
      if (m_ipv4->GetNInterfaces () > 2 || m_associated == false)
        {
          SendHello ();
          Time time = m_helloInterval+JITTER;
//          std::cout << "Total = " << time << "\n";
          m_helloTimer.Schedule (time);
        }
    }  
}

///
/// \brief Clears the routing table and frees the memory assigned to each one of its entries.
///
void
MeDeHaDtnRouting::Clear ()
{
  NS_LOG_FUNCTION_NOARGS ();
//  m_table.clear ();
}

Ptr<Ipv4Route>
MeDeHaDtnRouting::RouteOutput (Ptr<Packet> p, const Ipv4Header &header, Ptr<NetDevice> oif, Socket::SocketErrno &sockerr)
{  
  Ipv4Address destination = header.GetDestination ();
  Ptr<Ipv4Route> rtentry = 0;

  rtentry = LookupMeDeHaDtn (destination);
  if (rtentry)
    { 
      sockerr = Socket::ERROR_NOTERROR;
    }
  else
    { 
      sockerr = Socket::ERROR_NOROUTETOHOST;
    }
  return rtentry;
}

bool MeDeHaDtnRouting::RouteInput  (Ptr<const Packet> p, 
  const Ipv4Header &header, Ptr<const NetDevice> idev,
  UnicastForwardCallback ucb, MulticastForwardCallback mcb,             
  LocalDeliverCallback lcb, ErrorCallback ecb)
{   
  NS_LOG_FUNCTION (this << " " << m_ipv4->GetObject<Node> ()->GetId() << " " << header.GetDestination ());
//
// This is a unicast packet.  Check to see if we have a route for it.
//
  NS_LOG_LOGIC ("Unicast destination");
  Ptr<Ipv4Route> rtentry = LookupMeDeHaDtn (header.GetDestination ());
  if (rtentry != 0)
    {
      NS_LOG_LOGIC ("Found unicast destination- calling unicast callback");
      ucb (rtentry, p, header);  // unicast forwarding callback
      return true;
    }
  else
    {
      NS_LOG_LOGIC ("Did not find unicast destination- returning false");
      return false; // Let other routing protocols try to handle this
    }
}
void 
MeDeHaDtnRouting::NotifyInterfaceUp (uint32_t i)
{}
void 
MeDeHaDtnRouting::NotifyInterfaceDown (uint32_t i)
{}
void 
MeDeHaDtnRouting::NotifyAddAddress (uint32_t interface, Ipv4InterfaceAddress address)
{}
void 
MeDeHaDtnRouting::NotifyRemoveAddress (uint32_t interface, Ipv4InterfaceAddress address)
{}

void
MeDeHaDtnRouting::ConnectToCatchAssocEvent (void)
{
  m_node = m_ipv4->GetObject<Node> ();
  std::ostringstream oss;

  oss << "/NodeList/" << m_node->GetId () << "/DeviceList/*/Mac/$ns3::NqstaWifiMac/Assoc";
  Config::Connect (oss.str (), MakeCallback (&MeDeHaDtnRouting::ReceiveAssocNotif, this));

  std::ostringstream oss2;  
  oss2 << "/NodeList/" << m_node->GetId () << "/DeviceList/*/Mac/$ns3::NqstaWifiMac/DeAssoc";
  Config::Connect (oss2.str (), MakeCallback (&MeDeHaDtnRouting::ReceiveDisAssocNotif, this));
}


void
MeDeHaDtnRouting::ReceiveAssocNotif (std::string context, Mac48Address apAddress)
{
  Ptr<ArpL3Protocol> arp = m_ipv4->GetObject<ArpL3Protocol> ();
  
//  Ipv4Address ipv4AddressSta = arp->SearchForIpv4Address (staAddress);
  Ipv4Address ipv4AddressAP = arp->SearchForIpv4Address (apAddress);
  m_associatedWith = ipv4AddressAP;
  m_associated = true;

  uint32_t interface = 1;  
  uint16_t numOfIfces = m_ipv4->GetNInterfaces ();
  
  if (numOfIfces == 2)
    {
      interface = 1;
    }
  else 
    {
//      std::cout << "MeDeHaDtnRouting::ReceiveAssocNotif, main address is " << m_mainAddress << "\n";
      for (uint32_t i = 0; i < numOfIfces; i++)
        {
          Ipv4Address addr = m_ipv4->GetAddress (i, 0).GetLocal ();      
          Ipv4Address loopback ("127.0.0.1");
      
          if (addr != loopback &&
              addr != m_mainAddress)
            {
	      // Again, the assumption is that the second IP address
	      // (just after the ad hoc interface address)
	      // would be the address of Infrastructure interface
 	      interface = i;
//	      std::cout << "default interface = " << interface << " and address is " << addr << "\n";
	      break;
	    }
        }
    }
  
  SetDefaultRoute (ipv4AddressAP, interface);

#ifdef DEBUG
  std::cout << "MeDeHaDtnRouting::ReceiveAssocNotif (), Associated Event received: station mac = " << apAddress << ", station ip = " << ipv4AddressAP << "\n";
#endif
  
  Ptr<Ipv4L3Protocol> ipv4L3 = m_node->GetObject <Ipv4L3Protocol> ();
  HandleNotificationsClass hNotif = ipv4L3->GetHandleNotificationClass ();
  
  Ptr<Packet> packetToSend = Create<Packet> ();
  
  Ipv4Address addr[2];
  uint16_t j=0;
  if (numOfIfces == 2)
    {
      addr[0] = m_mainAddress;
    }
  else 
    {      
      // Here, we are assuming that if we have more than one interface, the second interface would always be Infrastructure interface
      for (uint32_t i = 0; i < m_ipv4->GetNInterfaces (); i++)
        {
	  Ipv4Address address = m_ipv4->GetAddress (i, 0).GetLocal ();	  
          Ipv4Address loopback ("127.0.0.1");
  
          if (address == loopback ||
              address == m_mainAddress)
            {
              continue;
	    }
	  else
	    {	      
              addr[j] = address;
	      j++;
	    }
        }
    }
    
  Time t = Simulator::Now ();

  Ipv4Address addresses[3];
  addresses[0] = addr[0];
  addresses[1] = m_mainAddress;
  addresses[2] = addr[1];
  
  hNotif.SendNotificationToAPs (STA_ASS_NOTIF, j+1, ipv4AddressAP, addresses, 0, packetToSend);
  ipv4L3->Send (packetToSend, m_mainAddress, m_associatedWith, 17, NULL);

#ifdef DEBUG  
  std::cout << "MeDeHaDtnRouting::ReceiveAssocNotif, buffer size at station " << addresses[0] << " is " << buffer.m_listSize << std::endl;
#endif

  if (buffer.m_listSize > 0)
    {
      if (m_txTimerSourceEvent.IsRunning ())
	      m_txTimerSourceEvent.Cancel ();
//      ForwardStoredFramesToAssociatedAPs (ipv4L3);
      m_txTimerSourceEvent = Simulator::Schedule (m_txTimerSource, &MeDeHaDtnRouting::HandleSourceStoredFrames, this);	// reschedule the timer for transmission of buffered frames
    }
      
}

void
MeDeHaDtnRouting::ReceiveDisAssocNotif (std::string context, Mac48Address apAddress)
{
  Ptr<ArpL3Protocol> arp = this->GetObject<ArpL3Protocol> ();
  
//  Ipv4Address ipv4AddressSta = arp->SearchForIpv4Address (staAddress);
  Ipv4Address ipv4AddressAP = arp->SearchForIpv4Address (apAddress);
  m_associated = false;
  m_associatedWith = "0.0.0.0";
  
//  m_niTimerEvent.Cancel ();
  m_defaultRoute = NULL;

#ifdef DEBUG
  std::cout << "MeDeHaDtnRouting::Disassociated Event received: station mac = " << apAddress << ", station ip = " << ipv4AddressAP << "\n";
#endif
}


void 
MeDeHaDtnRouting::AddHostRouteTo (Ipv4Address dest, 
                                   Ipv4Address nextHop, 
                                   uint32_t interface)
{
  NS_LOG_FUNCTION_NOARGS ();
  Ipv4RoutingTableEntry *route = new Ipv4RoutingTableEntry ();
  *route = Ipv4RoutingTableEntry::CreateHostRouteTo (dest, nextHop, interface);
  m_hostRoutes.push_back (route);
}

void 
MeDeHaDtnRouting::AddHostRouteTo (Ipv4Address dest, 
                                   uint32_t interface)
{
  NS_LOG_FUNCTION_NOARGS ();
  Ipv4RoutingTableEntry *route = new Ipv4RoutingTableEntry ();
  *route = Ipv4RoutingTableEntry::CreateHostRouteTo (dest, interface);
  m_hostRoutes.push_back (route);
}

void 
MeDeHaDtnRouting::SetDefaultRoute (Ipv4Address nextHop, 
                                    uint32_t interface)
{
  NS_LOG_FUNCTION_NOARGS ();
  Ipv4RoutingTableEntry *route = new Ipv4RoutingTableEntry ();
  *route = Ipv4RoutingTableEntry::CreateDefaultRoute (nextHop, interface);
  delete m_defaultRoute;
  m_defaultRoute = route;
}

uint32_t 
MeDeHaDtnRouting::GetNRoutes (void)
{
  NS_LOG_FUNCTION_NOARGS ();
  uint32_t n = 0;
  if (m_defaultRoute != 0)
    {
      n++;
    }
  n += m_hostRoutes.size ();
  return n;
}

Ipv4RoutingTableEntry
MeDeHaDtnRouting::GetDefaultRoute ()
{
  NS_LOG_FUNCTION_NOARGS ();
  if (m_defaultRoute != 0)
    {
      return *m_defaultRoute;
    }
  else
    {
      return Ipv4RoutingTableEntry ();
    }
}

Ipv4RoutingTableEntry 
MeDeHaDtnRouting::GetRoute (uint32_t index)
{
  NS_LOG_FUNCTION_NOARGS ();
  if (index == 0 && m_defaultRoute != 0)
    {
      return *m_defaultRoute;
    }
  if (index > 0 && m_defaultRoute != 0)
    {
      index--;
    }
  if (index < m_hostRoutes.size ())
    {
      uint32_t tmp = 0;
      for (HostRoutesCI i = m_hostRoutes.begin (); 
           i != m_hostRoutes.end (); 
           i++) 
        {
          if (tmp  == index)
            {
              return *i;
            }
          tmp++;
        }
    }
  NS_ASSERT (false);
  // quiet compiler.
  return 0;
}


void 
MeDeHaDtnRouting::RemoveRoute (uint32_t index)
{
  NS_LOG_FUNCTION_NOARGS ();
  if (index == 0 && m_defaultRoute != 0)
    {
      delete m_defaultRoute;
      m_defaultRoute = 0;
    }
  if (index > 0 && m_defaultRoute != 0)
    {
      index--;
    }
  if (index < m_hostRoutes.size ())
    {
      uint32_t tmp = 0;
      for (HostRoutesI i = m_hostRoutes.begin (); 
           i != m_hostRoutes.end (); 
           i++) 
        {
          if (tmp  == index)
            {
              delete *i;
              m_hostRoutes.erase (i);
              return;
            }
          tmp++;
        }
    }
  NS_ASSERT (false);
}


Ptr<Ipv4Route>
MeDeHaDtnRouting::LookupMeDeHaDtn (Ipv4Address dest)
{
  NS_LOG_FUNCTION_NOARGS ();

#ifdef DEBUG
  std::cout << "At DTN Node " << m_mainAddress << " MeDeHa route lookup is called\n";
#endif

  Ipv4Address outDest;
  bool status = GetIPAddressForDtnMode (dest, &outDest);
  
  Ptr<Ipv4Route> rtentry = 0;
  if (status)
    {
      for (HostRoutesCI i = m_hostRoutes.begin (); 
           i != m_hostRoutes.end (); 
           i++) 
        {
          NS_ASSERT ((*i)->IsHost ());
          if ((*i)->GetDest ().IsEqual (dest))
            {
              NS_LOG_LOGIC ("Found global host route" << *i);
              Ipv4RoutingTableEntry* route = (*i);
              rtentry = Create<Ipv4Route> ();
              uint32_t interfaceIdx = route->GetInterface ();
              rtentry->SetDestination (route->GetDest ());
              rtentry->SetSource (SourceAddressSelection (interfaceIdx, route->GetDest ()));
              rtentry->SetGateway (route->GetGateway ());
              rtentry->SetOutputDevice (m_ipv4->GetNetDevice (interfaceIdx));
#ifdef DEBUG
	      std::cout << Simulator::Now () << ", At DTN Node " << m_mainAddress << " MeDeHa DTN Route Lookup for " << dest << ", Gateway is " << route->GetGateway () << ", and Interface is " << route->GetInterface () << "\n";
#endif
	      NS_LOG_LOGIC ("At DTN Node " << m_mainAddress << " MeDeHa DTN Route Lookup for " << dest << ", Gateway is " << route->GetGateway () << ", and Interface is " << route->GetInterface ());
              return rtentry;
            }
        }
    }

  if (m_defaultRoute != 0) 
    {
      NS_ASSERT (m_defaultRoute->IsDefault ());
      NS_LOG_LOGIC ("Found global network route" << m_defaultRoute);
      Ipv4RoutingTableEntry* route = m_defaultRoute;
      rtentry = Create<Ipv4Route> ();
      uint32_t interfaceIdx = route->GetInterface ();
      rtentry->SetDestination (route->GetDest ());
      rtentry->SetSource (SourceAddressSelection (interfaceIdx, route->GetDest ()));
      rtentry->SetGateway (route->GetGateway ());
      rtentry->SetOutputDevice (m_ipv4->GetNetDevice (interfaceIdx));
#ifdef DEBUG
      std::cout << "MeDeHaDtnRouting::LookupMeDeHaDtn (), At " << m_mainAddress << ", Returning default route\n";
#endif
      return rtentry;
    }
  return 0;
}


Ipv4Address
MeDeHaDtnRouting::SourceAddressSelection (uint32_t interfaceIdx, Ipv4Address dest)
{
  if (m_ipv4->GetNAddresses (interfaceIdx) == 1)  // common case
    {
      return m_ipv4->GetAddress (interfaceIdx, 0).GetLocal ();
    }
  // no way to determine the scope of the destination, so adopt the
  // following rule:  pick the first available address (index 0) unless
  // a subsequent address is on link (in which case, pick the primary
  // address if there are multiple)
  Ipv4Address candidate = m_ipv4->GetAddress (interfaceIdx, 0).GetLocal ();
  for (uint32_t i = 0; i < m_ipv4->GetNAddresses (interfaceIdx); i++)
    {
      Ipv4InterfaceAddress test = m_ipv4->GetAddress (interfaceIdx, i);
      if (test.GetLocal ().CombineMask (test.GetMask ()) == dest.CombineMask (test.GetMask ()))
        {
          if (test.IsSecondary () == false) 
            {
              return test.GetLocal ();
            }
        }
    }
  return candidate;
}


bool
MeDeHaDtnRouting::RefreshEntryForDest (Ipv4Address destAddress, Ipv4Address nextHopAddr, Ipv4Address* addressArray, uint16_t length, bool iN)
{
  uint32_t interface = 0;
  for (uint32_t i = 0; i < m_ipv4->GetNInterfaces (); i++)
    {
      Ipv4Address addr = m_ipv4->GetAddress (i, 0).GetLocal ();
      if (addr == m_mainAddress)
        {
	  interface = i;
	  break;
	}
    }

  bool status = 0;

  if (iN)
    {
      status = RemovingRoutingEntry (destAddress);
      
#ifdef DEBUG
      std::cout << "MeDeHaDtnRouting::RefreshEntryForDest, Route is added for destination " << destAddress << "\n";
#endif

      AddHostRouteTo (destAddress, interface);  
    }
  else
    {
      bool result = SearchForDestEntryAsImmediateNeighbor (destAddress, addressArray, length);
      if (!result)
        {
          status = RemovingRoutingEntry (destAddress);
	  
#ifdef DEBUG
          std::cout << "MeDeHaDtnRouting::RefreshEntryForDest, Route is added for destination " << destAddress << "\n";
#endif

	  AddHostRouteTo (destAddress, nextHopAddr, interface);
	}
    }
  
  return status;
}
/*
void
MeDeHaDtnRouting::HandleNITimer (void)
{
//  std::cout << "MeDeHaDtnRouting::HandleNITimer, Timer is called\n";
  
  // Here, We are just sending only the routing table information one-by-one
  // TODO:: We need to send contact table information as well
  // If a station's information is found in the routing table, then we don't send its contact table information
  
  CheckForExpiredEntries ();
  if (m_associated)
    {
      Ptr<Ipv4L3Protocol> ipv4L3 = m_node->GetObject <Ipv4L3Protocol> ();
      HandleNotificationsClass hNotif = ipv4L3->GetHandleNotificationClass ();
      RTbleI iterator;
      
      Ipv4Address addr;
      for (uint32_t i = 0; i < m_ipv4->GetNInterfaces (); i++)
        {
          addr = m_ipv4->GetAddress (i, 0).GetLocal ();
          Ipv4Address loopback ("127.0.0.1");
      
          if (addr != loopback &&
              addr != m_mainAddress)
            { 	      
              break;
            }
        }
      
      for (iterator = m_routingTable.begin (); iterator != m_routingTable.end (); iterator++)
        {
          Ptr<Packet> packetToSend = Create<Packet> ();
 	    
	  // header.GetIPAddress01 () signifies the ad hoc interface address of a station
	  // while sending information to an associated AP, we keep the infrastructure address "header.GetIPAddress02 ()" of the station first
          hNotif.SendNotificationToAPs (NEIGHBORHOOD_INFO_NOTIF, 2, addr, iterator->staAddress, iterator->staAddressAlternate, packetToSend);
          ipv4L3->Send (packetToSend, m_mainAddress, m_associatedWith, 17, NULL);
	}
    }
    
  m_niTimerEvent = Simulator::Schedule (m_niInterval, &MeDeHaDtnRouting::HandleNITimer, this);
}
*/


void
MeDeHaDtnRouting::UpdateContactTable (Ipv4Address* address, uint16_t length, bool status)
{
  MeDeHaDtnRouting::ContactTable* ct;
  uint16_t i=0;
  
  while (i < length)
    {
      ct = SearchForDestEntryInContactTable (address [i]);
      if (ct != NULL)
      	  break;
      else
          i++;
    }

  if (ct != NULL)
    {
      ct->staAddress = address[0];
      
      if (length >= 2)
        {
          ct->staAltAddress01 = address[1];
        }
      if (length >= 3)
        {
          ct->staAltAddress02 = address[2];
        }

      ct->encounterTime = Simulator::Now ();
      if (status == 0)
        {
          ct->numOfEncounters++;
        }
      return;
    }
    
/*
  CTbleI iterator;
  for (iterator = m_contactTable.begin (); iterator != m_contactTable.end (); iterator++)
    {

      if ((iterator->staAddress == address[0] &&
           iterator->staAltAddress01 == address[1]) ||
          (iterator->staAddress == address[1] &&
           iterator->staAltAddress01 == address[0]))
	{
	  iterator->encounterTime = Simulator::Now ();
	  if (status == 0)
	    {
              iterator->numOfEncounters++;
	    }
	  return;
	}
    }
*/

  struct ContactTable ctToAdd;
  ctToAdd.staAddress = address[0];
  if (length >= 2)
    {
      ctToAdd.staAltAddress01 = address[1];
    }
  if (length >= 3)
    {
      ctToAdd.staAltAddress02 = address[2];
    }
    
  ctToAdd.encounterTime = Simulator::Now ();
  ctToAdd.numOfEncounters = 1;
  
  m_contactTable.push_back (ctToAdd);
}


void
MeDeHaDtnRouting::UpdateRoutingTable (Ipv4Address* address, uint16_t length, Ipv4Address nextHop, bool iN)
{
  RTbleI iterator;
  
  CheckForExpiredEntries ();

  for (iterator = m_routingTable.begin (); iterator != m_routingTable.end (); iterator++)
    {
      uint16_t i=0;
      while (i<length)
        {
//	  std::cout << iterator->staAddress << " " << iterator->staAltAddress01 << " " << iterator->staAltAddress02 << " " << address[i] << "\n";
          if (iterator->staAddress == address[i] ||
              iterator->staAltAddress01 == address[i] ||
              iterator->staAltAddress02 == address[i])
            {
//	      std::cout << "&&&&&&&&&&&&&&updating routing entry for " << address[0] << " in local routing table&&&&&&&&&&&&&&&&\n";
              iterator->staAddress = address[0];
              if (length >= 2)
                {
                  iterator->staAltAddress01 = address[1];
                }
              if (length >= 3)
                {
                  iterator->staAltAddress02 = address[2];
                }
              iterator->nextHop = nextHop;
              iterator->entryTime = Simulator::Now ();
              iterator->immediateNeighbor = iN;
              return;
            }
          i++;
        }
    }
    
/*  
  RTbleI iterator;
  
  CheckForExpiredEntries ();
  for (iterator = m_routingTable.begin (); iterator != m_routingTable.end (); iterator++)
    {
      if ((iterator->staAddress == address[0] &&
           iterator->staAltAddress01 == address[1]) ||
          (iterator->staAddress == address[1] &&
           iterator->staAltAddress01 == address[0]))
	{
	  iterator->nextHop = nextHop;
	  iterator->entryTime = Simulator::Now ();
	  iterator->immediateNeighbor = iN;
	  return;
	}
    }
*/

  struct RoutingTable rt;
  rt.staAddress = address[0];
  if (length >= 2)
    {
      rt.staAltAddress01 = address[1];
    }
  if (length >= 3)
    {
      rt.staAltAddress02 = address[2];
    }  
  
  rt.nextHop = nextHop;
  rt.entryTime = Simulator::Now ();
  rt.immediateNeighbor = iN;
  
  m_routingTable.push_back (rt);

/* To be removed
  for (iterator = m_routingTable.begin (); iterator != m_routingTable.end (); iterator++)
    {
      std::cout << "routing entry is " << iterator->staAddress << " address to add is " << address[0] << " and " << address[1] << " next hop is " << iterator->nextHop << "\n";
    }
*/
}


bool
MeDeHaDtnRouting::GetIPAddressForDtnMode (Ipv4Address inputAddress, Ipv4Address* outputAddress)
{
  RTbleI iterator;
  
  for (iterator = m_routingTable.begin (); iterator != m_routingTable.end (); iterator++)
    {
      if (iterator->staAddress == inputAddress)
        {
	  *outputAddress = inputAddress;
	  return true;
	}
      else if (iterator->staAltAddress01 == inputAddress ||
               iterator->staAltAddress02 == inputAddress)
        {
	  *outputAddress = iterator->staAddress;
	  return true;
	}
    }
  return false;
}


void
MeDeHaDtnRouting::CheckForExpiredEntries (void)
{
  Ipv4Address addr;
  for (uint32_t i = 0; i < m_ipv4->GetNInterfaces (); i++)
    {
      addr = m_ipv4->GetAddress (i, 0).GetLocal ();
      Ipv4Address loopback ("127.0.0.1");
      
      if (addr != loopback &&
          addr != m_mainAddress)
        {
	  // Again, here we assume that the 2nd address is IP address of the infrstructure interface
          m_addressAlternate01 = addr; 	      
	  break;
	}
    }

  RTbleI iterator;
  Time currentTime = Simulator::Now ();
  
  for (iterator = m_routingTable.begin (); iterator != m_routingTable.end (); )
    {
//	  std::cout << "CheckForExpiredEntries (), address 3 = " << iterator->staAddress << std::endl;
      if ((currentTime - (iterator->entryTime)) > ENTRY_EXPIRE_TIME)
        {
#ifdef DEBUG
          std::cout << "Calling from CheckForExpiredEntries () " << ENTRY_EXPIRE_TIME << " Entry Time = " << iterator->entryTime << " Current Time = " << currentTime << "\n";
#endif
          RemovingRoutingEntry (iterator->staAddress);
          RemovingRoutingEntry (iterator->staAltAddress01);
          RemovingRoutingEntry (iterator->staAltAddress02);

          Ipv4Address addresses[3];
          addresses[0] = iterator->staAddress;
          addresses[1] = iterator->staAltAddress01;
          addresses[2] = iterator->staAltAddress02;

//          std::cout << "CheckForExpiredEntries (), sender: " << addr << " first: " << iterator->staAddress << " second: " << iterator->staAltAddress01 << "\n";
   	      iterator = m_routingTable.erase (iterator);

          RemoveRouteWithGWAsDest (addresses[0]);

          if (m_associated)
            {
              Ptr<Ipv4L3Protocol> ipv4L3 = m_node->GetObject <Ipv4L3Protocol> ();
              HandleNotificationsClass hNotif = ipv4L3->GetHandleNotificationClass ();
              Ptr<Packet> packetToSend = Create<Packet> ();
              hNotif.SendNotificationToAPs (NEIGHBORHOOD_DISASS_NOTIF, 3, addr, addresses, 0, packetToSend);
              ipv4L3->Send (packetToSend, m_mainAddress, m_associatedWith, 17, NULL);
            }
          iterator = m_routingTable.begin ();
          continue;
        }
      else
        {
          iterator++;
        }
    }
}


bool
MeDeHaDtnRouting::RemovingRoutingEntry (Ipv4Address destAddress)
{
  bool status = 0;
  for (uint32_t i = 0; i < GetNRoutes (); i++)
    {
      Ipv4RoutingTableEntry route = GetRoute (i);
      if (route.GetDest () == destAddress)
        {
#ifdef DEBUG
	  std::cout << "MeDeHaDtnRouting::RemovingRoutingEntry, Removing Route for " << destAddress << ", Gateway is " << route.GetGateway () << ", and Interface is " << route.GetInterface () << "\n";
#endif
          RemoveRoute (i);
	  status = 1;
          break;
        }
    }
  return status;
}


bool
MeDeHaDtnRouting::SearchForDestEntryAsImmediateNeighbor (Ipv4Address destAddress, Ipv4Address* addressArray, uint16_t length)
{
  for (uint32_t i = 0; i < GetNRoutes (); i++)
    {
      Ipv4RoutingTableEntry route = GetRoute (i);
      
      if (route.GetDest () == destAddress)
        {	  
	  if (route.GetGateway () == Ipv4Address ("0.0.0.0"))
	    {
	      return 1;
	    }
	  else
	    {
              uint16_t l=0;
	      while (l<length)
	        {
		  if (route.GetGateway () == addressArray[l])
                    {
                      return 1;
                    }
		  l++;
		}
            }
        }
    }
  return 0;
}


bool
MeDeHaDtnRouting::SearchForDestEntryInRoutingTable (Ipv4Address destAddress)
{
  RTbleI iterator;
   
  if (!m_routingTable.empty ())
  {
  for (iterator = m_routingTable.begin (); iterator != m_routingTable.end (); iterator++)
    {
      if (iterator->staAddress == destAddress ||
          iterator->staAltAddress01 == destAddress ||
          iterator->staAltAddress02 == destAddress ||
          iterator->nextHop == destAddress)
        {
	  return true;
	}
    }
  }
  return false;
}


void
MeDeHaDtnRouting::MeDeHaBufferPacketLocally (Ptr<const Packet> packet, Ipv4Header ipHeader, UdpHeader* udpHeader)
{
  MedehaIntraManetNotifHeader hdr;
#ifdef BUFFER
  UdpHeader udpHdr;
//  Packet pkt = *packet;
  Ptr<Packet> pkt = packet->Copy ();

  if (ipHeader.GetProtocol () == 18)
    {
      std::cout << "Check Point 2\n";
      pkt->RemoveHeader (hdr);
    }

  if (udpHeader != 0)
    {
      udpHdr = *udpHeader;
    }
  else
    {
      pkt->RemoveHeader (udpHdr);
    }
  SeqNumHeader seqHdr;
  pkt->RemoveHeader (seqHdr);
  int seqNum = seqHdr.GetSeqNum ();
  int priority = seqHdr.GetPriority ();
  int copies = seqHdr.GetNumOfCopies ();
  bool copyGiven = seqHdr.GetCopyGiven ();
//  seqHdr.SetNumOfCopies (copies-1);
//  std::cout << "MeDeHaDtnRouting::MeDeHaBufferPacketLocally, Packet with Seq Num " << seqNum << " buffered\n";
  pkt->AddHeader (seqHdr);
  pkt->AddHeader (udpHdr);
  
  BufferedFramesSt frameForBuffer;
  frameForBuffer.packet = pkt; //ConstCast<Packet> (packet);
  frameForBuffer.src = ipHeader.GetSource ();

  if (ipHeader.GetProtocol () == 18)
    {
      std::cout << "Check Point 3\n"; 
      frameForBuffer.dst = hdr.GetDstIPAddress ();
      frameForBuffer.protocol = hdr.GetProtocol ();
    }
  else
    {
      frameForBuffer.dst = ipHeader.GetDestination ();
      frameForBuffer.protocol = ipHeader.GetProtocol ();
    }
  
  frameForBuffer.seqNum = seqNum;
  frameForBuffer.priority = priority;
  frameForBuffer.copies = copies;
  frameForBuffer.bbCopyGiven = copyGiven;
  frameForBuffer.ttl = 20;
  
  bool status = false;
  status = RemoveDuplicateCopiesFromBuffer (frameForBuffer);
  
  if (!status && copies > 0 && copies < MAX_NUM_OF_COPIES)
    {
      buffer.AddElementToList(frameForBuffer);
      std::cout << "MeDeHaDtnRouting::MeDeHaBufferPacketLocally, At " << m_mainAddress << ", buffering packet no. " << seqNum << " for " << ipHeader.GetDestination () << "\n";
    }
#endif
}


bool
MeDeHaDtnRouting::RemoveDuplicateCopiesFromBuffer (BufferedFramesSt frame)
{
  std::list<BufferedFramesSt>::iterator newIterator;		// link list iterator
  for (newIterator = buffer.m_pktList.begin(); newIterator != buffer.m_pktList.end(); )
    {
      if (newIterator->seqNum == frame.seqNum &&
          newIterator->src == frame.src &&
          newIterator->dst == frame.dst)
        {
	  frame.copies++;
          return true;
//          newIterator = buffer.m_pktList.erase (newIterator);
        }
      else
        {
          newIterator++;
        }
    }
  return false;
}


void
MeDeHaDtnRouting::SendStoredFrames (Ptr<Ipv4L3Protocol> ipv4L3)
{
  RTbleI iterator;  
  
  Ptr<Packet> packet;
  int counter = 0;
  int i = 0;
  Ipv4Address destAddress, alternateAddress01, alternateAddress02;
  
  std::cout << "At " << m_mainAddress << ", SendStoredFrames is Called\n";

  bool flag = 0;

  for (iterator = m_routingTable.begin (), i=0; ;)
    {
      destAddress = iterator->staAddress;      
      alternateAddress01 = iterator->staAltAddress01;
      alternateAddress02 = iterator->staAltAddress02;
//    std::cout << "SendStoreFramesToDest(): dest address = " << destAddress << " my address = " << myAddress << endl;		

      int count01 = CountNumberOfStoredFrames (destAddress);
      int count02 = CountNumberOfStoredFrames (alternateAddress01);
      int count03 = CountNumberOfStoredFrames (alternateAddress02);
      
//    std::cout << "SendStoreFramesToDest(): Count for dest " << destAddress << " is " << count << endl;

      if (count01 > 0)
        {
          std::list<BufferedFramesSt>::iterator newIterator;		// link list iterator
          for (newIterator = buffer.m_pktList.begin(); newIterator != buffer.m_pktList.end(); )
            {
              packet = (newIterator->packet);

              if (destAddress == newIterator->dst)
                {
                  flag = 1;
#ifdef DEBUG
		  std::cout << "SendStoredFrames (): Sending message to " << destAddress << "\n";
#endif
		  ipv4L3->Send (packet, newIterator->src, newIterator->dst, newIterator->protocol, NULL);
		  newIterator = buffer.m_pktList.erase(newIterator);
                  (buffer.m_listSize)--;
		  buffer.SetMemoryState ();
                  i++;
                  break;
                }
              else
                {
                  counter++;
                  newIterator++;
                }                   
            }
        }
	
      if (count02 > 0)
        {
          std::list<BufferedFramesSt>::iterator newIterator;		// link list iterator
          for (newIterator = buffer.m_pktList.begin(); newIterator != buffer.m_pktList.end(); )
            {
              packet = (newIterator->packet);

              if (alternateAddress01 == newIterator->dst)
                {
                  flag = 1;
#ifdef DEBUG
		  std::cout << "SendStoredFrames (): Sending message to " << alternateAddress01 << "\n";
#endif
		  ipv4L3->Send (packet, newIterator->src, newIterator->dst, newIterator->protocol, NULL);
		  newIterator = buffer.m_pktList.erase(newIterator);
                  (buffer.m_listSize)--;
		  buffer.SetMemoryState ();
                  i++;
                  break;
                }
              else
                {
                  counter++;
                  newIterator++;
                }
            }
        }
	
      if (count03 > 0)
        {
          std::list<BufferedFramesSt>::iterator newIterator;		// link list iterator
          for (newIterator = buffer.m_pktList.begin(); newIterator != buffer.m_pktList.end(); )
            {
              packet = (newIterator->packet);

              if (alternateAddress02 == newIterator->dst)
                {
                  flag = 1;
#ifdef DEBUG
		  std::cout << "SendStoredFrames (): Sending message to " << alternateAddress02 << "\n";
#endif
		  ipv4L3->Send (packet, newIterator->src, newIterator->dst, newIterator->protocol, NULL);
		  newIterator = buffer.m_pktList.erase(newIterator);
                  (buffer.m_listSize)--;
		  buffer.SetMemoryState ();
                  i++;
                  break;
                }
              else
                {
                  counter++;
                  newIterator++;
                }
            }
        }

	
      iterator++;		
      if (i == (MEDEHA_MAX_NUM_FRAMES_TO_SEND) || buffer.m_pktList.size() == 0)
        {				
          break;
        }
      if (iterator == m_routingTable.end())
        {
          if (flag)
            {
              iterator = m_routingTable.begin();
              flag = 0;
            }
          else
            {
              break;
            }
        }
    }    
}


int
MeDeHaDtnRouting::CountNumberOfStoredFrames (Ipv4Address destAddress)
{
  int counter = 0;
  
  std::list<BufferedFramesSt>::iterator newIterator;

  for (newIterator = buffer.m_pktList.begin(); newIterator != buffer.m_pktList.end(); newIterator++)
    {
      if (destAddress == newIterator->dst)
        {
          counter++;
        }      
    }
  return counter;
}


void
MeDeHaDtnRouting::HandleSendTimer (void)
{
  Ptr<Ipv4L3Protocol> ipv4L3 = m_node->GetObject <Ipv4L3Protocol> ();
  if (buffer.m_listSize > 0)
    {
      NS_LOG_INFO("Timer Expires, Sending stored frames");
      SendStoredFrames (ipv4L3);
    }
  if (buffer.m_listSize > 0)
    {
      m_txTimerEvent = Simulator::Schedule (m_txTimer, &MeDeHaDtnRouting::HandleSendTimer, this);	// reschedule the timer for transmission of buffered frames
    }
  else
    {
      m_txTimerEvent.Cancel ();
    }
}


MeDeHaDtnRouting::ContactTable*
MeDeHaDtnRouting::SearchForDestEntryInContactTable (Ipv4Address destAddress)
{
  CTbleI iterator;
  
  for (iterator = m_contactTable.begin (); iterator != m_contactTable.end (); iterator++)
    {
      if (iterator->staAddress == destAddress ||
          iterator->staAltAddress01 == destAddress ||
          iterator->staAltAddress02 == destAddress)
        {
          return &(*iterator);
        }
    }
  return NULL;
}


void
MeDeHaDtnRouting::HandleMultiCopyTimer (void)
{
  Ptr<Ipv4L3Protocol> ipv4L3 = m_node->GetObject <Ipv4L3Protocol> ();

  bool status = 0;
  if (buffer.m_listSize > 0)
    {
      NS_LOG_INFO("MeDeHaDtnRouting::HandleMultiCopyTimer, Timer Expires, Sending stored frames");
      status = CompareSeqNumAndSendPacket (ipv4L3, m_destMultiCopy01, m_destMultiCopy02, m_destMultiCopy03);
    }
    
  if (buffer.m_listSize > 0 && status == 1)
    {
      m_txTimerMultiCopyEvent = Simulator::Schedule (m_txTimerMultiCopy, &MeDeHaDtnRouting::HandleMultiCopyTimer, this);	// reschedule the timer for transmission of buffered frames
    }
    
  else
    {
      m_txTimerMultiCopyEvent.Cancel ();
    }
}


bool
MeDeHaDtnRouting::CompareSeqNumAndSendPacket (Ptr<Ipv4L3Protocol> ipv4L3, Ipv4Address destAddress01, Ipv4Address destAddress02, Ipv4Address destAddress03)
{
  uint16_t i=0;
  Ptr<Packet> packet;  
  std::list<BufferedFramesSt>::iterator newIterator;		// link list iterator
  
  for (newIterator = buffer.m_pktList.begin(); newIterator != buffer.m_pktList.end(); )
    {
      packet = (newIterator->packet);

      if (newIterator->dst == destAddress01 ||
          newIterator->dst == destAddress02 ||
          newIterator->dst == destAddress03)
        {
#ifdef DEBUG
          std::cout << "CompareSeqNumAndSendPacket (): Sending message to " << destAddress01 << "\n";
#endif
          bool status = CheckCommonStoredPacket (newIterator->src, newIterator->dst, newIterator->seqNum);
          if (!status)
            {
              Ptr<Packet> copy = packet->Copy ();
	  
              UdpHeader udp;
              SeqNumHeader seqHdr;
              packet->RemoveHeader (udp);
              packet->RemoveHeader (seqHdr);
              int copies = seqHdr.GetNumOfCopies ();
              seqHdr.SetNumOfCopies (1);
              packet->AddHeader (seqHdr);
              packet->AddHeader (udp);
	      
              ipv4L3->Send (packet, newIterator->src, newIterator->dst, newIterator->protocol, NULL);
              copies--;
	      
              if (copies > 0)
                {
                  copy->RemoveHeader (udp);
                  copy->RemoveHeader (seqHdr);
                  		  
                  seqHdr.SetNumOfCopies (copies);
                  std::cout << "CompareSeqNumAndSendPacket (), Remaining Copies for msg " << newIterator->seqNum << " = " << copies << std::endl;
                  copy->AddHeader (seqHdr);
                  copy->AddHeader (udp);
		  
                  newIterator->packet = copy;
               	  newIterator++;                  
              	}
              else
                {
                  newIterator = buffer.m_pktList.erase(newIterator);
                  (buffer.m_listSize)--;
                  buffer.SetMemoryState ();
              	}
              i++;
              if (i == 20)
                  break;
            }
          else
            {
              newIterator++;
            }                   
        }
      else
        {
          newIterator++;
        }
    }
    
  if (i == 0)
    {
    	m_msgVectorSummeryList.clear ();
      return 0;
    }
    
  return 1;
}


bool
MeDeHaDtnRouting::CheckCommonStoredPacket (Ipv4Address src, Ipv4Address dst, uint16_t seqNum)
{
  VSummeryI iter;
  
  for (iter=m_msgVectorSummeryList.begin (); iter!=m_msgVectorSummeryList.end (); iter++)
    {      
      if (iter->srcAddress == src &&
          iter->dstAddress == dst &&
          iter->seqNumber == seqNum)
        {
	  std::cout << "src: " << iter->srcAddress << ", " << src << " dst: " << iter->dstAddress << ", " << dst << " seqNum: " << iter->seqNumber << ", " << seqNum << "\n";
          return 1;
        }
    }
  return 0;
}


void
MeDeHaDtnRouting::HandleSourceStoredFrames (void)
{
  Ptr<Ipv4L3Protocol> ipv4L3 = m_node->GetObject <Ipv4L3Protocol> ();

  bool status = 0;
  if (buffer.m_listSize > 0)
    {
      NS_LOG_INFO("MeDeHaDtnRouting::HandleSourceStoredFrames, Timer Expires, Sending stored frames");
      status = ForwardStoredFramesToAssociatedAPs (ipv4L3);
    }
  if (buffer.m_listSize > 0 && status == 1)
    {
      m_txTimerSourceEvent = Simulator::Schedule (m_txTimerSource, &MeDeHaDtnRouting::HandleSourceStoredFrames, this);	// reschedule the timer for transmission of buffered frames
    }
  else
    {
      m_txTimerSourceEvent.Cancel ();
    }
}


bool
MeDeHaDtnRouting::ForwardStoredFramesToAssociatedAPs (Ptr<Ipv4L3Protocol> ipv4L3)
{
  Ptr<Packet> packet;  
  std::list<BufferedFramesSt>::iterator newIterator;		// link list iterator

#ifdef DEBUG
  std::cout << "MeDeHaDtnRouting::ForwardStoredFramesToAssociatedAPs, Forwarding frames to AP\n";
#endif

  int i=0;
  for (newIterator = buffer.m_pktList.begin(); newIterator != buffer.m_pktList.end(); )
    {
      packet = (newIterator->packet);

      if (!(newIterator->bbCopyGiven))
        {
#ifdef DEBUG
          std::cout << "MeDeHaDtnRouting::ForwardStoredFramesToAssociatedAPs, FLAG for dst " << newIterator->dst << " is " << newIterator->bbCopyGiven << std::endl;
#endif
#ifdef SOCIAL_AFFILIATION
	  bool flag = CompareSocialAffiliationWithAP (newIterator->dst, m_associatedWith);
	  if (flag)
	    {
#endif
	  Ptr<Packet> copy = packet->Copy ();
	  
          UdpHeader udp;
          SeqNumHeader seqHdr;
          packet->RemoveHeader (udp);
          packet->RemoveHeader (seqHdr);
          int copies = seqHdr.GetNumOfCopies ();
          seqHdr.SetNumOfCopies (1);
          packet->AddHeader (seqHdr);
          packet->AddHeader (udp);
          ipv4L3->Send (packet, newIterator->src, newIterator->dst, newIterator->protocol, NULL);
          
          copies--;
	  
          if (copies > 0)
            {
              copy->RemoveHeader (udp);
              copy->RemoveHeader (seqHdr);
              seqHdr.SetCopyGiven (true);
              seqHdr.SetNumOfCopies (copies);
              std::cout << "MeDeHaDtnRouting::ForwardStoredFramesToAssociatedAPs (), Remaining Copies for msg " << newIterator->seqNum << " = " << copies << std::endl;
              copy->AddHeader (seqHdr);
              copy->AddHeader (udp);

              newIterator->packet = copy;
              newIterator->bbCopyGiven = true;
              newIterator++;
	          }
          else
            {
              newIterator = buffer.m_pktList.erase (newIterator);
              (buffer.m_listSize)--;
              buffer.SetMemoryState ();
            }
	  
          i++;
          if (i == 10)
              break;
#ifdef SOCIAL_AFFILIATION
	    }
	  else
	    {
	      newIterator++;
	    }
#endif
        }
      else
        {
          newIterator++;
        }
    }
    
  if (i == 0)
    {
      return 0;
    }
    
  return 1;
}


void
MeDeHaDtnRouting::RemoveRouteWithGWAsDest (Ipv4Address gatewayAddress)
{
  Ipv4Address destAddress;
  
  for (uint32_t i = 0; i < GetNRoutes (); i++)
    {
      Ipv4RoutingTableEntry route = GetRoute (i);
      if (route.GetGateway () == gatewayAddress)
        {          
          destAddress = route.GetDest ();
          uint16_t status = RemoveFromRoutingTable (destAddress, gatewayAddress);
          
          if (status == 2)
            {
              continue;
            }
#ifdef DEBUG
          std::cout << "MeDeHaDtnRouting::RemoveRouteWithGWAsDest, Removing Route for " << route.GetDest () << ", Gateway is " << route.GetGateway () << ", and Interface is " << route.GetInterface () << "\n";
#endif
          RemoveRoute (i);
        }
    }
}


uint16_t
MeDeHaDtnRouting::RemoveFromRoutingTable (Ipv4Address destAddress, Ipv4Address gatewayAddress)
{
  RTbleI iterator;
   
  if (!m_routingTable.empty ())
    {
      for (iterator = m_routingTable.begin (); iterator != m_routingTable.end ();)
        {
	  if ((iterator->staAddress == destAddress &&
	      iterator->staAltAddress01 == gatewayAddress) ||
	      (iterator->staAddress == gatewayAddress &&
	       (iterator->staAltAddress01 == destAddress ||
	        iterator->staAltAddress02 == destAddress)))
	    {
	      return 2;
	    }
	      
          if (iterator->staAddress == destAddress ||
              iterator->staAltAddress01 == destAddress ||
	      iterator->staAltAddress02 == destAddress ||
              iterator->nextHop == destAddress)
            {
              iterator = m_routingTable.erase (iterator);
              return 1;
            }
          else
            {
              iterator++;
            }
        }
    }    
  return 0;
}


bool
MeDeHaDtnRouting::CompareSocialAffiliationWithAP (Ipv4Address dstAddress, 
					    Ipv4Address apAddress)
{
  uint32_t dst = dstAddress.Get () & 0xff;
  uint32_t relay = apAddress.Get () & 0xff;
  uint32_t compare = dstAddress.Get() >> 24;
  
  std::cout << "CompareSocialAffiliation (), " << dst << ", " << relay << ", " << compare << std::endl;
  
  if (compare == 0x0a)
    {
      dst -= 9;
    }

  std::cout << "CompareSocialAffiliation (), " << dstAddress << ", " << apAddress << std::endl;
  
  if (dst >= 1 && dst <= 14)
    {
      if (relay >= 1 && relay <= 3)
        {
          return true;
        }
    }
  else if (dst >= 15 && dst <= 28)
    {
      if (relay >= 4 && relay <= 6)
        {
          return true;
        }
    }
  else if (dst >= 29 && dst <= 42)
    {
      if (relay >= 7 && relay <= 9)
        {
          return true;
        }
    }
    
  return false;
}


bool
MeDeHaDtnRouting::CompareSocialAffiliation (Ipv4Address dstAddress, 
					    Ipv4Address relayAddress)
{
  uint32_t dst = dstAddress.Get () & 0xff;
  uint32_t relay = relayAddress.Get () & 0xff;
  uint32_t compare = dstAddress.Get() >> 24;
  
  std::cout << "CompareSocialAffiliation (), " << dst << ", " << relay << ", " << compare << std::endl;
  
  if (compare == 0x0a)
    {
      dst -= 9;
    }

  std::cout << "CompareSocialAffiliation (), " << dstAddress << ", " << relayAddress << std::endl;
  
  if (dst >= 1 && dst <= 14)
    {
      if (relay >= 43 && relay <= 48)
        {
          return true;
        }
    }
  else if (dst >= 15 && dst <= 28)
    {
      if (relay >= 49 && relay <= 54)
        {
          return true;
        }
    }
  else if (dst >= 29 && dst <= 42)
    {
      if (relay >= 55 && relay <= 60)
        {
          return true;
        }
    }
    
  return false;
}


void
MeDeHaDtnRouting::ForwardStoredFramesToSocialNode (Ptr<Ipv4L3Protocol> ipv4L3, Ipv4Address relayAddress)
{
  Ptr<Packet> packet;  
  std::list<BufferedFramesSt>::iterator newIterator;		// link list iterator
  uint16_t i=0;
  
  for (newIterator = buffer.m_pktList.begin(); newIterator != buffer.m_pktList.end(); )
    {
      packet = (newIterator->packet);
      bool status = CompareSocialAffiliation (newIterator->dst, relayAddress);
      
      if (status)
        {
          bool flag = CheckCommonStoredPacket (newIterator->src, newIterator->dst, newIterator->seqNum);
	  
          if (!flag)
            {
              std::cout << "Forwarding frame to social node " << relayAddress << "\n";
              UpdateRoutingTable (&newIterator->dst, 1, relayAddress, 0);
              RefreshEntryForDest (newIterator->dst, relayAddress, &newIterator->dst, 1, 0);
	  
              Ptr<Packet> copy = packet->Copy ();
	  
              UdpHeader udp;
              SeqNumHeader seqHdr;
              packet->RemoveHeader (udp);
              packet->RemoveHeader (seqHdr);
              int copies = seqHdr.GetNumOfCopies ();
              seqHdr.SetNumOfCopies (1);
              packet->AddHeader (seqHdr);
              packet->AddHeader (udp);

              ipv4L3->Send (packet, newIterator->src, newIterator->dst, newIterator->protocol, NULL);
              copies--;

              if (copies > 0)
                {
                  copy->RemoveHeader (udp);
                  copy->RemoveHeader (seqHdr);
                  		  
                  seqHdr.SetNumOfCopies (copies);
                  std::cout << "ForwardStoredFramesToSocialNode (), Remaining Copies for msg " << newIterator->seqNum << " = " << copies << std::endl;
                  copy->AddHeader (seqHdr);
                  copy->AddHeader (udp);
		  
                  newIterator->packet = copy;
                  newIterator++;                  
                }
              else
                {
                  newIterator = buffer.m_pktList.erase(newIterator);
                  (buffer.m_listSize)--;
                  buffer.SetMemoryState ();
                }
              i++;
              if (i == 20)
                  break;
            }
          else
            {
              newIterator++;
            }
        }
      else
        {
	  newIterator++;
        }
    }
}


void
MeDeHaDtnRouting::ConnectToCatchManetEvent (void)
{
  std::ostringstream oss;
  std::cout << "ConnectToCatchManetEvent " << m_node->GetId () << "\n";
  oss << "/NodeList/" << m_node->GetId () << "/$ns3::olsr::RoutingProtocol/ManetConnected";
  Config::Connect (oss.str (), MakeCallback (&MeDeHaDtnRouting::ReceiveManetNotif, this));

  std::ostringstream oss2;  
  oss2 << "/NodeList/" << m_node->GetId () << "/$ns3::olsr::RoutingProtocol/RoutingTableChanged";
  Config::Connect (oss2.str (), MakeCallback (&MeDeHaDtnRouting::UpdateManetRoutingTable, this));
}


void
MeDeHaDtnRouting::UpdateManetRoutingTable (std::string context, uint32_t size)
{
  int16_t priority1 = 10;
  Ptr<Ipv4L3Protocol> ipv4L3 = m_node->GetObject <Ipv4L3Protocol> ();
  Ptr<Ipv4ListRouting> ipv4LRP;
  ipv4LRP = DynamicCast<Ipv4ListRouting> (ipv4L3->GetRoutingProtocol ());
  Ptr<olsr::RoutingProtocol> olsrRouting =  DynamicCast<olsr::RoutingProtocol> (ipv4LRP->GetRoutingProtocol (0, priority1));
  
  if (olsrRouting == NULL)
    {
      return;
    }
  m_olsrRtEntries = olsrRouting->GetEntries ();
  UpdateListOfNonOlsrNodes ();
//  std::cout << "UpdateManetRoutingTable (): MANET routing table is updated\n";
 
//  std::vector<olsr::RoutingTableEntry>::iterator it; 
//  for (it = m_olsrRtEntries.begin (); it != m_olsrRtEntries.end (); it++)
//    {
//      uint16_t dest = (it->destAddr >> 8) & 0xff;
      
//      std::cout << "At Node " << m_mainAddress << ", UpdateManetRoutingTable (): Address is " << it->destAddr << "\n";
//    }

}


void
MeDeHaDtnRouting::ReceiveManetNotif (std::string context)
{
  Ipv4Address addr;
//  std::cout << "GetNInterfaces = " << m_ipv4->GetNInterfaces () << " main address = " << m_mainAddress << std::endl;
  for (uint32_t i = 0; i < m_ipv4->GetNInterfaces (); i++)
    {
      addr = m_ipv4->GetAddress (i, 0).GetLocal ();
//      std::cout << "address " << i << " is " << addr << "\n";
      Ipv4Address loopback ("127.0.0.1");
      
      if (addr != loopback &&
          addr != m_mainAddress)
        {
	  // Again, here we assume that the 2nd address is IP address of the infrstructure interface
          m_addressAlternate01 = addr; 	      
          break;
        }
    }

  static int counter = 0;
  if (m_manetTimerEvent.IsRunning ())
    {
      m_manetTimerEvent.Cancel ();      
    }

  m_manetTimerEvent = Simulator::Schedule (m_manetInterval, &MeDeHaDtnRouting::HandleManetTimer, this);
    
  counter++;
  if (!m_manet ||
      counter%50 == 0)
    {
      std::cout << "At Node " << m_mainAddress << " (" << m_node->GetId () << ") hello is received on MANET network \n";
  
      int16_t priority1 = 10;
    
      Ptr<Ipv4L3Protocol> ipv4L3 = m_node->GetObject <Ipv4L3Protocol> ();
      Ptr<Ipv4ListRouting> ipv4LRP;
      ipv4LRP = DynamicCast<Ipv4ListRouting> (ipv4L3->GetRoutingProtocol ());
      Ptr<olsr::RoutingProtocol> olsrRouting =  DynamicCast<olsr::RoutingProtocol> (ipv4LRP->GetRoutingProtocol (0, priority1));
  
      if (olsrRouting == NULL)
        {
          return;
        }
      m_olsrRtEntries = olsrRouting->GetEntries ();
      UpdateListOfNonOlsrNodes ();
      
      std::vector<olsr::RoutingTableEntry>::iterator it;
      for (it = m_olsrRtEntries.begin (); it != m_olsrRtEntries.end (); it++)
        {
//          std::cout << "Address is " << it->destAddr << "\n";
          if (!m_manet)
            {
              UpdateContactTable (&it->destAddr, 1, 0);
            }
          else
            {
              UpdateContactTable (&it->destAddr, 1, 1);
            }
        }
      
      if (buffer.GetBufferSize () > 0)
        {
          if (m_txTimerEventManet.IsRunning ())
            {
              m_txTimerEventManet.Cancel ();
            }
          SendStoredFramesToManet (ipv4L3);

#ifdef MEDEHA_MANET_PH_2
	  SendStoredFramesToOtherNetworkViaManet (ipv4L3);
#endif

          m_txTimerEventManet = Simulator::Schedule (m_txTimerManet, &MeDeHaDtnRouting::HandleSendTimerForManet, this);	// reschedule the timer for transmission of buffered frames
        }
	
      if (m_associated)
        {
          std::cout << "MeDeHaDtnRouting::ReceiveManetNotif (), At " << m_mainAddress << ", Manet is connected and station is associated\n";
          std::vector<olsr::RoutingTableEntry>::iterator it;
          for (it = m_olsrRtEntries.begin (); it != m_olsrRtEntries.end (); it++)
            {
              HandleNotificationsClass hNotif = ipv4L3->GetHandleNotificationClass ();
              Ptr<Packet> packetToSend = Create<Packet> ();
              hNotif.SendNotificationToAPs (NEIGHBORHOOD_INFO_NOTIF, 1, addr, &(it->destAddr), 1, packetToSend);
              ipv4L3->Send (packetToSend, m_mainAddress, m_associatedWith, 17, NULL);
            }
        }
      m_manet = true;
    }
}


void
MeDeHaDtnRouting::HandleManetTimer (void)
{
  std::cout << "************************* Disconnecting from MANET Network *************************\n";
  m_manetTimerEvent.Cancel ();
  m_manet = false;
  
  if (m_associated)
    {
      std::vector<olsr::RoutingTableEntry>::iterator it;
      for (it = m_olsrRtEntries.begin (); it != m_olsrRtEntries.end (); it++)
        {
//	  std::cout << "MeDeHaDtnRouting::HandleManetTimer (), Address is " << it->destAddr << "\n";
	}
    }
  
  m_olsrRtEntries.clear ();
  
  if (m_associated)
    {
      Ipv4Address addr;
      std::cout << "MeDeHaDtnRouting::HandleManetTimer (), GetNInterfaces = " << m_ipv4->GetNInterfaces () << " main address = " << m_mainAddress << std::endl;
      for (uint32_t i = 0; i < m_ipv4->GetNInterfaces (); i++)
        {
          addr = m_ipv4->GetAddress (i, 0).GetLocal ();
//          std::cout << "address " << i << " is " << addr << "\n";
          Ipv4Address loopback ("127.0.0.1");
      
          if (addr != loopback &&
              addr != m_mainAddress)
            {
	      // Again, here we assume that the 2nd address is IP address of the infrstructure interface
              m_addressAlternate01 = addr; 	      
	      break;
	    }
	}      
      
      Ptr<Ipv4L3Protocol> ipv4L3 = m_node->GetObject <Ipv4L3Protocol> ();
      HandleNotificationsClass hNotif = ipv4L3->GetHandleNotificationClass ();
      Ptr<Packet> packetToSend = Create<Packet> ();
      hNotif.SendNotificationToAPs (MANET_DISCONNECT_NOTIF, 1, addr, &addr, 1, packetToSend);
      ipv4L3->Send (packetToSend, m_mainAddress, m_associatedWith, 17, NULL);
    }
}


void
MeDeHaDtnRouting::ProcessManetPeerInfo (MeDeHaDtnMessageHeader msg, Ipv4Address senderAddress)
{
  MeDeHaDtnMessageHeader::ManetPeerInfo manetPI = msg.GetManetPeerInfo ();
  
#ifdef DEBUG
  std::cout << "\nDTN node: " << m_mainAddress << ", MANET_PEER_INFO message is received from " << senderAddress << " and contains:\n";
  std::cout << "Number of Stations = " << (int) msg.GetNumOfStations () << "\n";
#endif

  std::vector<MeDeHaDtnMessageHeader::Interfaces>::iterator iterIfces = manetPI.m_interfaces.begin ();
}


bool
MeDeHaDtnRouting::SearchForDestEntryInManetRoutingTable (Ipv4Address* destAddress, uint16_t len)
{
  std::vector<olsr::RoutingTableEntry>::iterator it;
  for (it = m_olsrRtEntries.begin (); it != m_olsrRtEntries.end (); it++)
    {
      uint16_t i=0;
      while (i<len)
        {
	  if (it->destAddr == destAddress[i])
            {
              return true;        
            }
	  i++;
        }
    }
    
  return false;
}


void
MeDeHaDtnRouting::HandleSendTimerForManet (void)
{
  Ptr<Ipv4L3Protocol> ipv4L3 = m_node->GetObject <Ipv4L3Protocol> ();
  if (buffer.m_listSize > 0)
    {
      NS_LOG_INFO("Timer Expires, Sending stored frames");
      SendStoredFramesToManet (ipv4L3);
//      SendStoredFramesToOtherNetworkViaManet (ipv4L3);
    }
  if (buffer.m_listSize > 0)
    {
      m_txTimerEventManet = Simulator::Schedule (m_txTimerManet, &MeDeHaDtnRouting::HandleSendTimerForManet, this);	// reschedule the timer for transmission of buffered frames
    }
  else
    {
      m_txTimerEventManet.Cancel ();
    }
}


void
MeDeHaDtnRouting::SendStoredFramesToManet (Ptr<Ipv4L3Protocol> ipv4L3)
{
  std::vector<olsr::RoutingTableEntry>::iterator it;
  
  Ptr<Packet> packet;
  int counter = 0;
  int i = 0;
  Ipv4Address destAddress;
  
//  std::cout << "At " << m_mainAddress << ", SendStoredFramesToManet is Called\n";

  bool flag = 0;

  for (it = m_olsrRtEntries.begin (), i=0; it != m_olsrRtEntries.end ();)
    {
      destAddress = it->destAddr;

      int count01 = CountNumberOfStoredFrames (destAddress);

      if (count01 > 0)
        {
          std::list<BufferedFramesSt>::iterator newIterator;		// link list iterator
          for (newIterator = buffer.m_pktList.begin(); newIterator != buffer.m_pktList.end(); )
            {
              packet = (newIterator->packet);

              if (destAddress == newIterator->dst)
                {
                  flag = 1;
#ifdef DEBUG
                  std::cout << "SendStoredFramesToManet (): Sending message to " << destAddress << "\n";
#endif
                  ipv4L3->Send (packet, newIterator->src, newIterator->dst, newIterator->protocol, NULL);
                  newIterator = buffer.m_pktList.erase(newIterator);
                  (buffer.m_listSize)--;
                  buffer.SetMemoryState ();
                  i++;
                  break;
                }              
              else
                {
                  counter++;
                  newIterator++;
                }                   
            }
        }
/*	
      if (flag)
        {
          i++;
        }
*/	
      it++;
      if (i == (MEDEHA_MAX_NUM_FRAMES_TO_SEND) || buffer.m_pktList.size() == 0)
        {				
          break;
        }
      if (it == m_olsrRtEntries.end())
        {
          if (flag)
            {
              it = m_olsrRtEntries.begin();
              flag = 0;
            }
          else
            {
              break;
            }
        }
    }    
}



void
MeDeHaDtnRouting::SendStoredFramesToOtherNetworkViaManet (Ptr<Ipv4L3Protocol> ipv4L3)
{
  std::vector<Ipv4Address>::iterator it;
  
  Ptr<Packet> packet;
  int counter = 0;
  int i = 0;
  Ipv4Address destAddress;
  
//  std::cout << "At " << m_mainAddress << ", SendStoredFramesToManet is Called\n";

  bool flag = 0;

  for (it = m_nonOlsrRtEntries.begin (), i=0; it != m_nonOlsrRtEntries.end ();)
    {
      destAddress = *it;
//      std::cout << "Setting tempDstAddress\n";
      Ipv4Address tempDstAddress = destAddress;
      
      if (tempDstAddress == "192.168.1.19" ||
	      tempDstAddress == "192.168.1.20" ||
	      tempDstAddress == "192.168.2.19" ||
	      tempDstAddress == "192.168.2.20" ||
	      tempDstAddress == "192.168.3.19" ||
	      tempDstAddress == "192.168.3.20")
        {

          std::list<BufferedFramesSt>::iterator newIterator;		// link list iterator
          for (newIterator = buffer.m_pktList.begin(); newIterator != buffer.m_pktList.end(); )
            {	  
              packet = (newIterator->packet);	  

              if (((tempDstAddress.Get () >> 8) & 0xff) == ((newIterator->dst.Get () >> 8) & 0xff) &&
                  ((tempDstAddress.Get () >> 8) & 0xff) != ((m_mainAddress.Get () >> 8) & 0xff) &&
                  ((tempDstAddress.Get () >> 8) & 0xff) != ((m_addressAlternate01.Get () >> 8) & 0xff))
                {
                  flag = 1;
//#ifdef DEBUG
                  std::cout << i << " ---> SendStoredFramesToOtherNetworkViaManet (): Sending message to " << tempDstAddress << " " << newIterator->dst << " " << m_mainAddress << " " << m_addressAlternate01 << " " << m_addressAlternate02 << "\n";
//#endif
                  MedehaIntraManetNotifHeader hdr;
                  hdr.SetType (0xA0);
                  hdr.SetProtocol (newIterator->protocol);
                  hdr.SetDstIPAddress (newIterator->dst);

                  Ptr<Packet> copy = packet->Copy ();

                  UdpHeader udp;
                  SeqNumHeader seqHdr;
                  packet->RemoveHeader (udp);
                  packet->RemoveHeader (seqHdr);
                  int copies = seqHdr.GetNumOfCopies ();
                  seqHdr.SetNumOfCopies (1);
                  packet->AddHeader (seqHdr);
                  packet->AddHeader (udp);	      
                  packet->AddHeader (hdr);
	      
                  ipv4L3->Send (packet, newIterator->src, tempDstAddress, 18, NULL);
	      
                  std::cout << "copies = " << copies << std::endl;
                  copies--;

                  if (copies > 0 && copies < MAX_NUM_OF_COPIES)
                    {
                      copy->RemoveHeader (udp);
                      copy->RemoveHeader (seqHdr);

                      seqHdr.SetNumOfCopies (copies);
//                      std::cout << "ForwardStoredFramesToSocialNode (), Remaining Copies for msg " << newIterator->seqNum << " = " << copies << std::endl;
                      copy->AddHeader (seqHdr);
                      copy->AddHeader (udp);

                      newIterator->packet = copy;
                      newIterator++;
                    }
                  else
                    {
                      newIterator = buffer.m_pktList.erase(newIterator);
                      (buffer.m_listSize)--;
                      buffer.SetMemoryState ();
                    }

                  i++;
                  if (i%5 == 0)
	            {
                      std::cout << "SendStoredFramesToOtherNetworkViaManet (), breaking the inner loop at " << i << "\n";
                      break;
                    }
                }
              else
                {
//	          std::cout << "Not Matched\n";
                  counter++;
                  newIterator++;
                }          
            }
        }
      it++;
      if (i >= 20 || buffer.m_pktList.size() == 0)
        {				
          break;
        }
      if (it == m_nonOlsrRtEntries.end())
        {
          if (flag)
            {
              it = m_nonOlsrRtEntries.begin();
              flag = 0;
            }
          else
            {
              break;
            }
        }
    }
}



bool
MeDeHaDtnRouting::GetManet (void)
{
  return m_manet;
}


uint16_t
MeDeHaDtnRouting::GetMeDeHaRoutingCode (void)
{
  return m_routingProtocolCode;	
}


Ipv4Address
MeDeHaDtnRouting::GetManetAddressFromAdhoc (Ipv4Address nodeAddress)
{
  RTbleI iterator;
  Ipv4Address manetAddress;

  for (iterator = m_routingTable.begin (); iterator != m_routingTable.end (); iterator++)
    {
      if (iterator->staAddress == nodeAddress)
        {
    	  manetAddress = iterator->staAltAddress02;
    	  break;
        }
    }
  return manetAddress;
}


void
MeDeHaDtnRouting::ForwardStoredFramesToManetSocialNode (Ptr<Ipv4L3Protocol> ipv4L3, 
                                                        Ipv4Address relayAddress,
                                                        Ipv4Address nextHop)
{
  Ptr<Packet> packet;
  std::list<BufferedFramesSt>::iterator newIterator;		// link list iterator
  uint16_t i=0;

  for (newIterator = buffer.m_pktList.begin(); newIterator != buffer.m_pktList.end(); )
    {
      packet = (newIterator->packet);
      bool status = CompareSocialAffiliationForManet (newIterator->dst, relayAddress);

      if (status)
        {
          bool flag = CheckCommonStoredPacket (newIterator->src, newIterator->dst, newIterator->seqNum);

          if (!flag)
            {
              std::cout << "ForwardStoredFramesToManetSocialNode (), Forwarding frame to social node " << relayAddress << "\n";
              UpdateRoutingTable (&newIterator->dst, 1, nextHop, 0);
              RefreshEntryForDest (newIterator->dst, nextHop, &newIterator->dst, 1, 0);

              Ptr<Packet> copy = packet->Copy ();

              UdpHeader udp;
              SeqNumHeader seqHdr;
              packet->RemoveHeader (udp);
              packet->RemoveHeader (seqHdr);
              int copies = seqHdr.GetNumOfCopies ();
              seqHdr.SetNumOfCopies (1);
              packet->AddHeader (seqHdr);
              packet->AddHeader (udp);

              ipv4L3->Send (packet, newIterator->src, newIterator->dst, newIterator->protocol, NULL);
              copies--;

              if (copies > 0)
                {
                  copy->RemoveHeader (udp);
                  copy->RemoveHeader (seqHdr);

                  seqHdr.SetNumOfCopies (copies);
//                  std::cout << "ForwardStoredFramesToSocialNode (), Remaining Copies for msg " << newIterator->seqNum << " = " << copies << std::endl;
                  copy->AddHeader (seqHdr);
                  copy->AddHeader (udp);

                  newIterator->packet = copy;
                  newIterator++;
                }
              else
                {
                  newIterator = buffer.m_pktList.erase(newIterator);
                  (buffer.m_listSize)--;
                  buffer.SetMemoryState ();
                }
              i++;
//              RemovingRoutingEntry (newIterator->dst);
//              RemoveFromRoutingTable (newIterator->dst, nextHop);
              if (i == 20)
                  break;
            }
          else
            {
              newIterator++;
            }
        }
      else
        {
          newIterator++;
        }
    }
}


bool
MeDeHaDtnRouting::CompareSocialAffiliationForManet (Ipv4Address dstAddress,
                                                    Ipv4Address relayAddress)
{
//  uint32_t dst = dstAddress.Get () & 0xff;
//  uint32_t relay = relayAddress.Get () & 0xff;
  uint32_t compare = (dstAddress.Get() >> 8) & 0xff;
  uint32_t relay = (relayAddress.Get () >> 8) & 0xff;

 // std::cout << "CompareSocialAffiliationForManet (), " << relay << ", " << compare << std::endl;
/*
  if (compare == 0x0a)
    {
      dst -= 9;
    }
*/
//  std::cout << "CompareSocialAffiliationForManet (), " << dstAddress << ", " << relayAddress << std::endl;

  if (relay == compare)
	  return true;
  else
	  return false;
/*
  if (dst >= 1 && dst <= 14)
    {
      if (relay >= 43 && relay <= 48)
        {
          return true;
        }
    }
  else if (dst >= 15 && dst <= 28)
    {
      if (relay >= 49 && relay <= 54)
        {
          return true;
        }
    }
  else if (dst >= 29 && dst <= 42)
    {
      if (relay >= 55 && relay <= 60)
        {
          return true;
        }
    }
*/
  return false;
}


void
MeDeHaDtnRouting::UpdateListOfNonOlsrNodes (void)
{
  m_nonOlsrRtEntries.clear ();
  
  std::vector<olsr::RoutingTableEntry>::iterator it;

  for (it = m_olsrRtEntries.begin (); it != m_olsrRtEntries.end (); it++)
    {
      Ipv4Address destAddress = it->destAddr;
      
      uint16_t addr = (destAddress.Get () >> 8) & 0xff;
      if (addr != 7 &&
          addr != 8 &&
	  addr != 9)
        {
          m_nonOlsrRtEntries.push_back (destAddress);
//	  std::cout << "UpdateListOfNonOlsrNodes(), At " << m_mainAddress << " Adding Entry " << destAddress << " addr = " << addr << std::endl;
	}
    }
}


}
